"""Contains DataBuffer class."""

import numpy as np
import pandas as pd
import casadi as cs
import torch

from typing import Optional, List, Union, Any, Type, Iterable, Dict
from regelum.typing import RgArray, RgArrayType
from .batch_sampler import RollingBatchSampler, BatchSampler
from collections import defaultdict
from ..optimizable import OptimizerConfig


class DataBuffer:
    """DataBuffer class for storing run data.

    DataBuffer is a container for storing run data: observations, actions,
    running objectives, iteration ids, episode ids, step ids. It is designed to store any
    data of numeric format.
    """

    def __init__(
        self,
    ):
        """Instantiate a DataBuffer."""
        self.nullify_buffer()

    def delete_key(self, key) -> None:
        self.data.pop(key)

    def keys(self) -> List[str]:
        return list(self.data.keys())

    def nullify_buffer(self) -> None:
        self.data = defaultdict(list)
        self.keys_for_indexing = None
        self.dtype_for_indexing = None
        self.device_for_indexing = None
        self.fill_na_for_indexing = None

    def update(self, data_in_dict_format: dict[str, RgArray]) -> None:
        for key, data_for_key in data_in_dict_format.items():
            self.data[key] = data_for_key

    def push_to_end(self, **kwargs) -> None:
        current_keys = set(self.data.keys())
        kwarg_keys = set(kwargs.keys())

        for _, data_item_for_key in kwargs.items():
            if np.any(np.isnan(data_item_for_key)):
                raise ValueError(
                    f"{type(data_item_for_key)} nan values are not allowed for `push_to_end` in data buffer"
                )
        is_line_added = False
        for key in current_keys.intersection(kwarg_keys):
            datum = np.array(kwargs[key])
            if np.any(np.isnan(self.data[key][-1])):
                self.data[key][-1] = datum
            else:
                self.data[key].append(datum)
                is_line_added = True

        buffer_len = len(self)
        for key in kwarg_keys.difference(current_keys):
            datum = np.array(kwargs[key])
            for _ in range(buffer_len - 1):
                self.data[key].append(np.full_like(datum, np.nan, dtype=float))
            self.data[key].append(datum)

        # if buffer len has changed fill all the rest keys with nan
        if is_line_added:
            for key in current_keys.difference(kwarg_keys):
                self.data[key].append(
                    np.full_like(self.data[key][-1], np.nan, dtype=float)
                )

    def last(self) -> dict[str, RgArray]:
        return self[-1]

    def to_dict(self):
        return self.data

    def to_pandas(
        self,
        keys: Optional[Union[List[str], Dict[str, Type[Any]]]] = None,
    ) -> pd.DataFrame:
        if isinstance(keys, list):
            _keys = {k: float for k in keys}
        elif isinstance(keys, dict):
            assert set(keys.keys()).issubset(
                self.data.keys()
            ), "keys must be a subset of data keys"
            _keys = keys
        elif keys is None:
            _keys = {k: float for k in self.data.keys()}
        else:
            raise AssertionError("keys must be a list or a dict or None")

        return pd.DataFrame(
            {
                k: (
                    np.array(self.data[k], dtype=_keys[k]).reshape(-1)
                    if self.data[k][0].size == 1
                    else self.data[k]
                )
                for k in _keys
            }
        )

    def __len__(self):
        if len(self.data.keys()) == 0:
            return 0
        else:
            return max([len(self.data[k]) for k in self.data.keys()])

    def _fill_na(self, arr: np.array, fill_na: Optional[float] = None) -> np.array:
        if fill_na is None:
            return arr
        else:
            np.nan_to_num(arr, copy=False, nan=fill_na)
            return arr

    def getitem(
        self,
        idx: Union[int, slice, Any],
        keys: Optional[Union[List[str], np.array]] = None,
        dtype: RgArrayType = np.array,
        device: Optional[Union[str, torch.device]] = None,
        fill_na: Optional[float] = 0.0,
    ) -> dict[str, RgArray]:
        _keys = keys if keys is not None else self.data.keys()
        difference_keys_set = set(_keys).difference(self.data.keys())
        if len(difference_keys_set) > 0:
            raise ValueError(
                f"Unexpected keys in DataBuffer.getitem: {difference_keys_set}"
            )

        if (
            isinstance(idx, int)
            or isinstance(idx, slice)
            or isinstance(idx, np.ndarray)
        ):
            if dtype == np.array:
                return {
                    key: self._fill_na(np.vstack(self.data[key])[idx], fill_na=fill_na)
                    for key in _keys
                }
            elif (
                dtype == torch.tensor
                or dtype == torch.FloatTensor
                or dtype == torch.DoubleTensor
                or dtype == torch.LongTensor
            ):
                if device is not None:
                    return {
                        key: dtype(
                            self._fill_na(np.vstack(self.data[key]), fill_na=fill_na)
                        )[idx].to(device)
                        for key in _keys
                    }
                else:
                    return {
                        key: dtype(
                            self._fill_na(np.vstack(self.data[key]), fill_na=fill_na)
                        )[idx]
                        for key in _keys
                    }
            elif dtype == cs.DM:
                return {
                    key: dtype(
                        self._fill_na(np.vstack(self.data[key]), fill_na=fill_na)[idx]
                    )
                    for key in _keys
                }
            else:
                raise ValueError(f"Unexpeted dtype in data_buffer.getitem: {dtype}")

    def set_indexing_rules(
        self,
        keys: List[str],
        dtype: RgArrayType,
        device: Optional[Union[str, torch.device]] = None,
        fill_na: Optional[float] = 0.0,
    ) -> None:
        self.keys_for_indexing = keys
        self.dtype_for_indexing = dtype
        self.device_for_indexing = device
        self.fill_na_for_indexing = fill_na

    def __getitem__(self, idx) -> dict[str, RgArray]:
        return self.getitem(
            idx,
            keys=self.keys_for_indexing,
            dtype=self.dtype_for_indexing,
            device=self.device_for_indexing,
            fill_na=self.fill_na_for_indexing,
        )

    def concat(self, other):
        for key in other.keys():
            self.data[key] += other.data[key]

    def get_latest(self, key: str) -> np.array:
        return self.data[key][-1]

    def iter_batches(
        self,
        keys: List[str],
        batch_sampler: Type[BatchSampler] = RollingBatchSampler,
        **batch_sampler_kwargs,
    ) -> Iterable[RgArray]:
        return batch_sampler(data_buffer=self, keys=keys, **batch_sampler_kwargs)

    def sample_last(
        self, keys: List[str], dtype: RgArrayType, n_samples: int = 1
    ) -> dict[str, RgArray]:
        if n_samples <= len(self):
            return self.getitem(
                np.arange(-n_samples, 0, dtype=int), keys=keys, dtype=dtype
            )

    def get_optimization_kwargs(
        self, keys: List[str], optimizer_config: OptimizerConfig
    ):
        config_options = optimizer_config.config_options

        method_name = config_options.get("data_buffer_sampling_method")
        kwargs = config_options.get("data_buffer_sampling_kwargs")
        assert (
            method_name is not None
        ), "Specify `data_buffer_sampling_method` in your optimizer_config"
        assert (
            kwargs is not None
        ), "Specify `data_buffer_sampling_kwargs` in your optimizer_config"

        if method_name == "iter_batches":
            return {"batch_sampler": self.iter_batches(keys=keys, **kwargs)}
        elif method_name == "sample_last":
            return self.sample_last(keys=keys, **kwargs)
        else:
            raise ValueError("Unknown data_buffer_sampling_method")



"""Contains models.

These can be used in system dynamics fitting, critic and other tasks.

Updates to come.

"""

from copy import deepcopy
from scipy.stats import truncnorm
import regelum


from .utils import rg

try:
    import torch
    import torch.nn.functional as F
    from torch import nn
    from torch.distributions.multivariate_normal import MultivariateNormal
    from torch.distributions.normal import Normal
except ModuleNotFoundError:
    from unittest.mock import MagicMock

    torch = MagicMock()
    nn = MagicMock()
    F = MagicMock()
    MultivariateNormal = MagicMock()

from abc import ABC, abstractmethod
from typing import Optional, Union, List, Any, Tuple, Dict, Callable
import numpy as np
import casadi as cs
from omegaconf.listconfig import ListConfig
import math


def force_positive_def(func):
    def positive_def_wrapper(obj, *args, **kwargs):
        if obj.force_positive_def:
            return rg.soft_abs(func(obj, *args, **kwargs))
        else:
            return func(obj, *args, **kwargs)

    return positive_def_wrapper


def unversal_model_call(obj, *argin, weights=None, use_stored_weights=False):
    """_summary_

    Args:
        obj: an object from which the function is called
        weights: if passed, the model will try to perform forward with passed weights. Defaults to None.
        use_stored_weights: If True, computes forward using cached model (for Torch always with detached weights). Defaults to False.

    Returns:
        An array or scalar, depends on the model definition
    """
    if len(argin) == 2:
        left, right = argin
        if len(left.shape) != len(right.shape):
            raise ValueError(
                f"In {obj.__class__.__name__}.__call__ left and right arguments must have same number of dimensions!"
            )

        dim = len(left.shape)

        if dim == 1:
            argin = rg.concatenate(argin, axis=0)
        elif dim == 2:
            argin = rg.concatenate(argin, axis=1)
        else:
            raise ValueError(
                f"Wrong number of dimensions in {obj.__class__.__name__}.__call__"
            )
    elif len(argin) == 1:
        argin = argin[0]
    else:
        raise ValueError(
            f"Wrong number of arguments in {obj.__class__.__name__}.__call__. Can be either 1 or 2. Got: {len(argin)}"
        )

    if use_stored_weights is False:
        if weights is not None:
            result = obj.forward(argin, weights)
        else:
            result = obj.forward(argin)
    else:
        result = obj.cache.forward(argin)

    return result


class Model(regelum.RegelumBase, ABC):
    """Blueprint of a model."""

    def __call__(self, *argin, weights=None, use_stored_weights=False):
        return unversal_model_call(
            self, *argin, weights=weights, use_stored_weights=use_stored_weights
        )

    @property
    def weights(self):
        return self._weights

    @property
    def named_parameters(self):
        return self.weights

    @weights.setter
    def weights(self, new_weights):
        assert (
            self.weights.shape == new_weights.shape
        ), "The shape of weights was changed "
        f"in runtime from {self.weights.shape} to {new_weights.shape}"

        self._weights = new_weights

    @abstractmethod
    def __init__(self):
        """Initialize an instance of a model."""
        pass

    @abstractmethod
    def forward(self):
        pass

    def update_weights(self, weights):
        self.weights = weights

    def cache_weights(self, weights=None):
        if "cache" not in self.__dict__.keys():
            self.cache = deepcopy(self)

        if weights is None:
            self.cache.update_weights(self.weights)
        else:
            self.cache.update_weights(weights)

    def update_and_cache_weights(self, weights):
        self.cache_weights(weights)
        self.update_weights(weights)

    def restore_weights(self):
        """Assign the weights of the cached model to the active model.

        This may be needed when pytorch optimizer resulted in unsatisfactory weights, for instance.

        """
        self.update_and_cache_weights(self.cache.weights)


class ModelQuadLin(Model):
    """Base class for generic quadratic linear models.

    Normally used for running objective specification (diagonal quadratic matrix without linear terms) and critic polynomial models.
    """

    def __init__(
        self,
        quad_matrix_type: str,
        is_with_linear_terms: bool = False,
        dim_inputs: int = None,
        weights: np.array = None,
        weight_min: float = 1.0e-6,
        weight_max: float = 1.0e3,
        add_random_init_noise: bool = False,
    ):
        """Initialize an instance of quadratic-linear model.

        Args:
            quad_matrix_type (str): Type of quadratic matrix. Can be
                'diagonal', 'full' or 'symmetric'.
            is_with_linear_terms (bool, optional): Whether include
                linear terms or not, defaults to False
            dim_inputs (int, optional): Dimension of system's (agent's)
                inputs, defaults to None
            weights (_type_, optional): Manual set of model weights,
                defaults to None
            weight_min (float, optional): Lower bound for weights,
                defaults to 1.0e-6
            weight_max (float, optional): Upper bound for weights,
                defaults to 1.0e3
        """
        assert (
            dim_inputs is not None or weights is not None
        ), "Need dim_inputs or weights"

        self.quad_matrix_type = quad_matrix_type
        self.is_with_linear_terms = is_with_linear_terms
        self.single_weight_min = weight_min
        self.single_weight_max = weight_max

        if weights is None:
            self._calculate_dims(dim_inputs)
            self.weight_min = weight_min * rg.ones(self.dim_weights)
            self.weight_max = weight_max * rg.ones(self.dim_weights)
            self.weights = (self.weight_min + self.weight_max) / 20.0
        else:
            self._calculate_dims(self._calculate_dim_inputs(len(weights)))
            assert self.dim_weights == len(weights), "Wrong shape of dim_weights"
            self.weights = rg.array(weights)

        self.add_random_init_noise = add_random_init_noise
        if self.add_random_init_noise:
            self.weight_min = np.random.uniform(
                self.weight_min, self.weight_max, self.dim_weights
            )
            self.weight_max = np.random.uniform(
                self.weight_min, self.weight_max, self.dim_weights
            )
            self.weights = np.random.uniform(
                self.weight_min, self.weight_max, self.dim_weights
            )

        self.cache_weights(self.weights)

    def get_quad_lin(self, new_weights):
        if self.quad_matrix_type == "full":
            quad_matrix = rg.reshape(
                new_weights[: self.dim_quad], (self.dim_inputs, self.dim_inputs)
            )
        elif self.quad_matrix_type == "diagonal":
            quad_matrix = rg.diag(new_weights[: self.dim_quad])
        elif self.quad_matrix_type == "symmetric":
            quad_matrix = ModelQuadLin.quad_matrix_from_flat_weights(
                new_weights[: self.dim_quad]
            )

        linear_coefs = (
            new_weights[self.dim_quad :] if self.is_with_linear_terms else None
        )

        return quad_matrix, linear_coefs

    @Model.weights.setter
    def weights(self, new_weights):
        if new_weights is not None:
            self._weights = new_weights
            self._quad_matrix, self._linear_coefs = self.get_quad_lin(new_weights)

    @property
    def weight_bounds(self):
        return rg.array([[self.single_weight_min, self.single_weight_max]])

    def _calculate_dim_inputs(self, dim_weights):
        if self.quad_matrix_type == "diagonal":
            if self.is_with_linear_terms:
                return dim_weights // 2
            else:
                return dim_weights
        elif self.quad_matrix_type == "full":
            if self.is_with_linear_terms:
                return round((np.sqrt(1 + 4 * dim_weights) - 1) / 2)
            else:
                return round(np.sqrt(dim_weights))
        elif self.quad_matrix_type == "symmetric":
            if self.is_with_linear_terms:
                return round((np.sqrt(9 + 8 * dim_weights) - 1) / 2)
            else:
                return round((np.sqrt(1 + 8 * dim_weights) - 1) / 2)

    def _calculate_dims(self, dim_inputs):
        self.dim_inputs = dim_inputs
        self.dim_linear = dim_inputs if self.is_with_linear_terms else 0
        if self.quad_matrix_type == "diagonal":
            self.dim_quad = dim_inputs
        elif self.quad_matrix_type == "full":
            self.dim_quad = dim_inputs * dim_inputs
        elif self.quad_matrix_type == "symmetric":
            self.dim_quad = dim_inputs * (dim_inputs + 1) // 2

        self.dim_weights = self.dim_quad + self.dim_linear

    def cast_to_inputs_type(self, value, inputs):
        if value is None:
            return None
        if isinstance(inputs, torch.Tensor):
            device = inputs.device
            if not isinstance(value, torch.Tensor):
                return torch.FloatTensor(value).to(device)
            elif device != value.device:
                return value.to(device)
        if isinstance(inputs, cs.MX) and isinstance(value, np.ndarray):
            value = rg.DM(value)
        return value

    def forward(self, inputs, weights=None):
        if weights is None:
            weights = self.weights
            quad_matrix = self._quad_matrix
            linear_coefs = self._linear_coefs
            if isinstance(inputs, torch.Tensor):
                quad_matrix = torch.FloatTensor(quad_matrix).to(inputs.device)
                if linear_coefs is not None:
                    linear_coefs = torch.FloatTensor(linear_coefs).to(inputs.device)
        else:
            quad_matrix, linear_coefs = self.get_quad_lin(weights)

        return ModelQuadLin.quadratic_linear_form(
            inputs,
            quad_matrix,
            linear_coefs,
        )

    @staticmethod
    def quad_matrix_from_flat_weights(
        flat_weights: Union[np.array, cs.DM, torch.Tensor], tol=1e-7
    ):
        len_flat_weights = flat_weights.shape[0]
        dim_quad_matrix_float = (np.sqrt(1 + 8 * len_flat_weights) - 1) / 2
        dim_quad_matrix = round(dim_quad_matrix_float)
        assert np.isclose(
            dim_quad_matrix_float, dim_quad_matrix, tol
        ), f"Can't build quad matrix with flat_weights of dim {len_flat_weights}"

        quad_matrix = rg.zeros(
            (dim_quad_matrix, dim_quad_matrix), prototype=flat_weights
        )
        left_ids, right_ids = np.triu_indices(dim_quad_matrix)
        for weigth_idx, (i, j) in enumerate(zip(left_ids, right_ids)):
            quad_matrix[i, j] = flat_weights[weigth_idx]

        return quad_matrix

    @staticmethod
    def quadratic_linear_form(inputs, quad_matrix, linear_coefs=None):
        initial_dim_inputs = len(inputs.shape)
        assert (
            initial_dim_inputs == 1 or initial_dim_inputs == 2
        ), "Wrong shape of inputs can be 1d or 2d. Got {}".format(initial_dim_inputs)

        if initial_dim_inputs == 1:
            inputs = inputs.reshape(1, -1)
        assert (
            len(quad_matrix.shape) == 2
        ), "Wrong shape of quad matrix. Should be 2d. Got{}".format(
            len(quad_matrix.shape)
        )
        assert (
            quad_matrix.shape[0] == quad_matrix.shape[1]
        ), "Quad matrix should be square"
        assert (
            quad_matrix.shape[0] == inputs.shape[1]
        ), "Quad matrix should have same number of rows as inputs"

        quadratic_term = inputs @ quad_matrix @ inputs.T
        if len(quadratic_term.shape) > 0:
            quadratic_term = rg.diag(quadratic_term)

        if linear_coefs is not None:
            # assert (
            #     len(linear_coefs.shape) == 2 and linear_coefs.shape[0] == 1
            # ), "Wrong shape of linear coefs. Should be (1,n). Got {}".format(
            #     linear_coefs.shape
            # )

            # assert (
            #     quad_matrix.shape[1] == linear_coefs.shape[1]
            # ), "Quad matrix should have same number of columns as linear coefs"

            linear_term = inputs @ linear_coefs
            output = quadratic_term + linear_term
        else:
            output = quadratic_term

        if initial_dim_inputs == 1:
            output = output.reshape(-1)
        return output


class ModelWeightContainer(Model):
    """Trivial model, which is typically used in policy in which actions are being optimized directly."""

    def __init__(self, dim_output: int, weights_init: Optional[np.array] = None):
        """Initialize an instance of a model returns weights on call independent of input.

        Args:
            dim_output: output dimension
            weights_init: initial weights to set
        """
        self.dim_output = dim_output
        self._weights = weights_init
        self.weights_init = weights_init
        self.update_and_cache_weights(self._weights)

    def forward(self, *argin, weights=None):
        if weights is not None:
            return rg.force_row(weights[0, : self.dim_output])
        else:
            return rg.force_row(self._weights[0, : self.dim_output])


class ModelNN(nn.Module):
    """Class of pytorch neural network models. This class is not to be used barebones.

    Instead, you should inherit from it and specify your concrete architecture.

    """

    @property
    def cache(self):
        """Isolate parameters of cached model from the current model."""
        return self.cached_model[0]

    def detach_weights(self):
        """Excludes the model's weights from the pytorch computation graph.

        This is needed to exclude the weights from the decision variables in optimization problems.
        An example is temporal-difference optimization, where the old critic is to be treated as a frozen model.

        """
        for variable in self.parameters():
            variable.detach_()

    def cache_weights(self, whatever=None):
        """Assign the active model weights to the cached model followed by a detach.

        This method also backs up itself and performs this operation only once upon the initialization procedure
        """
        if "cached_model" not in self.__dict__.keys():
            self.cached_model = (
                deepcopy(self),
            )  ## this is needed to prevent cached_model's parameters to be parsed by Torch module init hooks

        self.cache.load_state_dict(self.state_dict())
        self.cache.detach_weights()

    @property
    def weights(self):
        return self.state_dict()

    def update_weights(self, whatever=None):
        pass

    def update_and_cache_weights(self, weights=None):
        if weights is not None:
            for item in weights:
                weights[item].requires_grad_()
            weights = self.load_state_dict(weights)
        self.cache_weights()

    def restore_weights(self):
        """Assign the weights of the cached model to the active model.

        This may be needed when pytorch optimizer resulted in unsatisfactory weights, for instance.

        """
        self.update_and_cache_weights(self.cache.state_dict())

    def __call__(self, *argin, weights=None, use_stored_weights=False):
        return unversal_model_call(
            self, *argin, weights=weights, use_stored_weights=use_stored_weights
        )


class WeightClipper:
    """Weight clipper for pytorch layers."""

    def __init__(
        self, weight_min: Optional[float] = None, weight_max: Optional[float] = None
    ):
        """Initialize a weight clipper.

        Args:
            weight_min: minimum value for weight
            weight_max: maximum value for weight
        """
        self.weight_min = weight_min
        self.weight_max = weight_max

    def __call__(self, module):
        if self.weight_min is not None or self.weight_max is not None:
            # filter the variables to get the ones you want
            module.weight.data.clamp_(self.weight_min, self.weight_max)


class ModelPerceptron(ModelNN):
    """Helper class to ease the creation of perceptron models."""

    def __init__(
        self,
        dim_input: int,
        dim_output: int,
        dim_hidden: int,
        n_hidden_layers: int,
        hidden_activation: Optional[nn.Module] = None,
        output_activation: Optional[nn.Module] = None,
        force_positive_def: bool = False,
        is_force_infinitesimal: bool = False,
        is_bias: bool = True,
        weight_max: Optional[float] = None,
        weight_min: Optional[float] = None,
        linear_weights_init: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        linear_weights_init_kwargs: Optional[Dict[str, Any]] = None,
        biases_init: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        biases_init_kwargs: Optional[Dict[str, Any]] = None,
        output_bounds: Union[List[List[float]], np.array] = None,
        weights=None,
    ):
        """Initialize an instance of a fully-connected model.

        Args:
            dim_input (int): The dimensionality of the input.
            dim_output (int): The dimensionality of the output.
            dim_hidden (int): The dimensionality of the hidden linear
                layers (dim_hidden * dim_hidden).
            n_hidden_layers (int): The number of hidden layers.
            hidden_activation (Optional[nn.Module]): The activation
                function for the hidden layers. (Optional, defaults to
                None). If None then nn.LeakyReLU(0.2) is used.
            output_activation (Optional[nn.Module]): The activation
                function for the output layer. (Optional, defaults to
                None)
            force_positive_def (bool): Whether to force perceptron to be
                positive definite (if True rg.softabs is applied to output). (Optional,
                defaults to False)
            is_force_infinitesimal (bool): Whether to force perceptron
                to be equal to 0 with zero input. (Optional, defaults to
                False)
            is_bias (bool): Whether to include bias terms. (Optional,
                defaults to True)
            weight_max (Optional[float]): The maximum value for the
                weights. (Optional, defaults to None)
            weight_min (Optional[float]): The minimum value for the
                weights. (Optional, defaults to None)
            linear_weights_init (Optional[Callable[[torch.Tensor], torch.Tensor]]):
                The weight initialization function for the linear
                layers. (Optional, defaults to None)
            linear_weights_init_kwargs (Optional[Dict[str, Any]]):
                Additional keyword arguments for the linear weight
                initialization function. (Optional, defaults to None)
            biases_init (Optional[Callable[[torch.Tensor], torch.Tensor]]):
                The bias initialization function for the linear layers.
                (Optional, defaults to None)
            biases_init_kwargs (Optional[Dict[str, Any]]): Additional
                keyword arguments for the bias initialization function.
                (Optional, defaults to None)
            output_bounds (Union[List[List[float]], np.array]): The
                bounds for the output values. (Optional, defaults to
                None)
            weights (Optional): Pre-trained weights for the model.
                (Optional, defaults to None)
        """
        ModelNN.__init__(self)
        self.weight_clipper = WeightClipper(weight_min, weight_max)
        self.dim_input = dim_input
        self.dim_output = dim_output
        self.n_hidden_layers = n_hidden_layers
        self.force_positive_def = force_positive_def
        self.is_force_infinitesimal = is_force_infinitesimal
        self.is_bias = is_bias
        if isinstance(dim_hidden, int):
            self.input_layer = nn.Linear(dim_input, dim_hidden, bias=is_bias)
            self.hidden_layers = nn.ModuleList(
                [
                    nn.Linear(dim_hidden, dim_hidden, bias=is_bias)
                    for _ in range(n_hidden_layers)
                ]
            )
        elif isinstance(dim_hidden, list) or isinstance(dim_hidden, ListConfig):
            assert (
                len(dim_hidden) == n_hidden_layers
            ), "number of passed hidden dimensions is not equal to n_hidden_layers"
            self.input_layer = nn.Linear(dim_input, dim_hidden[0], bias=is_bias)
            dim_hidden_full = zip(dim_hidden[:-1], dim_hidden[1:])
            self.hidden_layers = nn.ModuleList(
                [
                    nn.Linear(dim_hidden_first, dim_hidden_second, bias=is_bias)
                    for dim_hidden_first, dim_hidden_second in dim_hidden_full
                ]
            )
            dim_hidden = dim_hidden[-1]

        else:
            raise ValueError("inappropriate type of dim_hidden argument passed.")
        self.output_layer = nn.Linear(dim_hidden, dim_output, bias=is_bias)
        self.hidden_activation = (
            hidden_activation if hidden_activation is not None else nn.LeakyReLU(0.15)
        )
        self.output_activation = output_activation
        self.bounds_handler = (
            BoundsHandler(output_bounds) if output_bounds is not None else None
        )
        if linear_weights_init is not None or biases_init is not None:
            for layer in (
                [self.input_layer]
                + [hidden_layer for hidden_layer in self.hidden_layers]
                + [self.output_layer]
            ):
                if linear_weights_init is not None:
                    if linear_weights_init_kwargs is not None:
                        linear_weights_init(layer.weight, **linear_weights_init_kwargs)
                    else:
                        linear_weights_init(layer.weight)

                if biases_init is not None:
                    if biases_init_kwargs is not None:
                        biases_init(layer.bias, **biases_init_kwargs)
                    else:
                        biases_init(layer.bias)

        if weights is not None:
            self.load_state_dict(weights)

        self.cache_weights()

    def _forward(self, x: torch.FloatTensor) -> torch.FloatTensor:
        self.input_layer.apply(self.weight_clipper)
        x = self.hidden_activation(self.input_layer(x))
        for hidden_layer in self.hidden_layers:
            hidden_layer.apply(self.weight_clipper)
            x = self.hidden_activation(hidden_layer(x))
        x = self.output_layer(x)
        if self.output_activation is not None:
            x = self.output_activation(x)

        if self.bounds_handler is not None:
            x = self.bounds_handler.unscale_from_minus_one_one_to_bounds(torch.tanh(x))

        return x

    @force_positive_def
    def forward(
        self, input_tensor: torch.FloatTensor, weights=None
    ) -> torch.FloatTensor:
        if self.is_force_infinitesimal:
            return self._forward(input_tensor) - self._forward(
                torch.zeros_like(input_tensor)
            )

        return self._forward(input_tensor)


class ModelWeightContainerTorch(ModelNN):
    """Pytorch model that have raw weights as its parameter.

    When using output bounds, the `output_bounding_type` parameter determines how the output is bounded:
    - If `output_bounding_type` is set to "clip", the output is clipped to the specified bounds.
    - If `output_bounding_type` is set to "tanh", the output is scaled using the hyperbolic tangent function to fit within the specified bounds. (same as squash_output in stable baselines)
    """

    def __init__(
        self,
        dim_weights: Union[int, Tuple[int, int]],
        output_bounds: Optional[List[Any]] = None,
        output_bounding_type: str = "clip",
    ):
        """Instantiate ModelWeightContainerTorch.

        Args:
            dim_weights (Union[int, Tuple[int, int]]): The
                dimensionality of the weights.
            output_bounds (Optional[List[Any]]): Optional bounds of the
                output. If `None`, the output is not bounded. Defaults
                to None.
            output_bounding_type (str): The type of output bounding.
                Must be either "clip" or "tanh". Defaults to "clip".
        """
        assert (
            output_bounding_type == "clip" or output_bounding_type == "tanh"
        ), "output_bounding_type must be 'clip' or 'tanh'"

        ModelNN.__init__(self)
        self.bounds_handler = (
            BoundsHandler(output_bounds) if output_bounds is not None else None
        )
        self.output_bounding_type = output_bounding_type

        self.dim_weights = (
            (1, dim_weights) if isinstance(dim_weights, int) else dim_weights
        )
        self._weights = torch.nn.Parameter(
            torch.FloatTensor(torch.zeros(self.dim_weights)),
            requires_grad=True,
        )

        self.cache_weights()

    def forward(self, inputs, weights=None):
        if self.bounds_handler is not None:
            if self.output_bounding_type == "clip":
                with torch.no_grad():
                    self._weights.clip_(-1.0, 1.0)

        if len(inputs.shape) == 1:
            inputs_like = self._weights[0, :]
        elif len(inputs.shape) == 2:
            if inputs.shape[0] <= self.dim_weights[0]:
                inputs_like = self._weights[: inputs.shape[0], :]
            else:
                raise ValueError(
                    "ModelWeightContainerTorch: Wrong inputs shape! inputs.shape[0]"
                    f"(Got: {inputs.shape[0]}) should be <= dim_weights[0] (Got: {self.dim_weights[0]})."
                )
        else:
            raise ValueError("Wrong inputs shape! Can be either 1 or 2")

        if self.bounds_handler is not None:
            if self.output_bounding_type == "clip":
                # inputs_like are already clipped in the beggining of the function via WeightClipper
                return self.bounds_handler.unscale_from_minus_one_one_to_bounds(
                    inputs_like
                )
            elif self.output_bounding_type == "tanh":
                return self.bounds_handler.unscale_from_minus_one_one_to_bounds(
                    torch.tanh(inputs_like)
                )
        else:
            return inputs_like


class BoundsHandler(ModelNN):
    r"""Output layer that restricts the model's output within specified bounds. The formula used is: F^{-1}(y), where F represents a linear transformation from the given bounds to the range [-1, 1].

    It is recommended to stack this layer after applying the tanh activation function to the model's output.
    """

    def __init__(self, bounds: Union[List[List[float]], np.array], is_unscale=True):
        """Initialize a new instance of the BoundsHandler class.

        Args:
            bounds (Union[List[List[float]], np.array] The bounds should be provided as a 2-column array-like object. The first column represents the left bounds, and the second column represents the right bounds.):
                Bounds of the model's output.
            is_unscale (bool, optional): Flag indicating whether to
                unscale (from [-1, 1] to bounds) or scale (from bounds
                to [-1, 1]) in forward. Defaults to True.
        """
        ModelNN.__init__(self)
        self.is_unscale = is_unscale
        self.bounds = torch.nn.Parameter(
            torch.FloatTensor(bounds),
            requires_grad=False,
        )

    def forward(self, x):
        if self.is_unscale:
            return self.unscale_from_minus_one_one_to_bounds(x)
        else:
            return self.scale_from_bounds_to_minus_one_one(x)

    def get_unscale_coefs_from_minus_one_one_to_bounds(
        self,
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        unscale_bias, unscale_multiplier = (
            self.bounds.mean(dim=1),
            (self.bounds[:, 1] - self.bounds[:, 0]) / 2.0,
        )
        return unscale_bias, unscale_multiplier

    def unscale_from_minus_one_one_to_bounds(
        self, x: torch.FloatTensor
    ) -> torch.FloatTensor:
        (
            unscale_bias,
            unscale_multiplier,
        ) = self.get_unscale_coefs_from_minus_one_one_to_bounds()

        return x * unscale_multiplier + unscale_bias

    def scale_from_bounds_to_minus_one_one(self, y):
        (
            unscale_bias,
            unscale_multiplier,
        ) = self.get_unscale_coefs_from_minus_one_one_to_bounds()

        return (y - unscale_bias) / unscale_multiplier


class MultiplyByConstant(nn.Module):
    """Represents a module that multiplies the input tensor by a constant value."""

    def __init__(self, constant: float) -> None:
        """Instatiate MultiplyByConstant.

        Args:
            constant (float): The constant value to multiply the input
                by.

        Returns:
            torch.Tensor: The tensor resulting from multiplying the
            input by the constant value.
        """
        super().__init__()
        self.constant = constant

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input * self.constant


class PerceptronWithTruncatedNormalNoise(ModelPerceptron):
    """Represents a perceptron model that applies a forward pass to input data and adds normal noise to the output.

    The `PerceptronWithTruncatedNormalNoise` class provides the following functionality:

    - Sampling : After the forward pass, normal noise is added to the output of the perceptron. The standard deviations of the noise can be specified using the `stds` parameter. All this functionality is implemented in the forward method.
    - Truncation to output bounds: The `is_truncated_to_output_bounds` flag determines whether the noise should be truncated to the output bounds. When set to `True`, the noise values are generated from corresponding truncated normal distribution.
    """

    def __init__(
        self,
        dim_input: int,
        dim_output: int,
        dim_hidden: int,
        n_hidden_layers: int,
        stds: Union[List[float], np.array] = None,
        hidden_activation: Optional[nn.Module] = None,
        output_activation: Optional[nn.Module] = None,
        force_positive_def: bool = False,
        is_force_infinitesimal: bool = False,
        is_bias: bool = True,
        weight_max: Optional[float] = None,
        weight_min: Optional[float] = None,
        linear_weights_init: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        linear_weights_init_kwargs: Optional[Dict[str, Any]] = None,
        biases_init: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        biases_init_kwargs: Optional[Dict[str, Any]] = None,
        output_bounds: Union[List[List[float]], np.array] = None,
        is_truncated_to_output_bounds: Optional[Dict[str, Any]] = False,
        weights=None,
    ):
        """Instantiate PerceptronWithTruncatedNormalNoise.

        Args:
            dim_input (int): The dimensionality of the input.
            dim_output (int): The dimensionality of the output.
            dim_hidden (int): The dimensionality of the hidden linear
                layers (dim_hidden * dim_hidden).
            n_hidden_layers (int): The number of hidden layers.
            hidden_activation (Optional[nn.Module]): The activation
                function for the hidden layers. If None,
                nn.LeakyReLU(0.2) is used. (Optional, defaults to None)
            output_activation (Optional[nn.Module]): The activation
                function for the output layer. (Optional, defaults to
                None)
            force_positive_def (bool): Whether to force the perceptron
                to be positive definite (if True, softabs is used).
                (Optional, defaults to False)
            is_force_infinitesimal (bool): Whether to force the
                perceptron to be equal to 0 with zero input. (Optional,
                defaults to False)
            is_bias (bool): Whether to include bias terms. (Optional,
                defaults to True)
            weight_max (Optional[float]): The maximum value for the
                weights. (Optional, defaults to None)
            weight_min (Optional[float]): The minimum value for the
                weights. (Optional, defaults to None)
            linear_weights_init (Optional[Callable[[torch.Tensor], torch.Tensor]]):
                The weight initialization function for the linear
                layers. (Optional, defaults to None)
            linear_weights_init_kwargs (Optional[Dict[str, Any]]):
                Additional keyword arguments for the linear weight
                initialization function. (Optional, defaults to None)
            biases_init (Optional[Callable[[torch.Tensor], torch.Tensor]]):
                The bias initialization function for the linear layers.
                (Optional, defaults to None)
            biases_init_kwargs (Optional[Dict[str, Any]]): Additional
                keyword arguments for the bias initialization function.
                (Optional, defaults to None)
            output_bounds (Union[List[List[float]], np.array]): The
                bounds for the output values. (Optional, defaults to
                None)
            weights (Optional): Pre-trained weights for the model.
                (Optional, defaults to None)
            stds (Optional[np.array]): The standard deviations for
                sampling the normal distribution. (Optional, defaults to
                None)
            is_truncated_to_output_bounds (bool): Whether to truncate
                the samples to the output bounds. (Optional, defaults to
                False)
        """
        super().__init__(
            dim_input=dim_input,
            dim_output=dim_output,
            dim_hidden=dim_hidden,
            n_hidden_layers=n_hidden_layers,
            hidden_activation=hidden_activation,
            output_activation=output_activation,
            force_positive_def=force_positive_def,
            is_force_infinitesimal=is_force_infinitesimal,
            is_bias=is_bias,
            weight_min=weight_min,
            weight_max=weight_max,
            output_bounds=output_bounds,
            linear_weights_init=linear_weights_init,
            linear_weights_init_kwargs=linear_weights_init_kwargs,
            biases_init=biases_init,
            biases_init_kwargs=biases_init_kwargs,
            weights=weights,
        )
        self.is_truncated_to_output_bounds = is_truncated_to_output_bounds
        self.output_bounds_array = np.array(output_bounds)
        if self.is_truncated_to_output_bounds:
            if output_bounds is None:
                raise AssertionError(
                    "Cannot truncate the output without specifying output bounds. Please set output_bounds and use is_truncated_to_output_bounds=True."
                )

        self.stds_array = np.array(stds).reshape(-1)
        assert (
            len(self.stds_array) == dim_output
        ), "The length of standard deviations should be equal to dim_output."
        self.stds = torch.nn.Parameter(
            torch.FloatTensor(self.stds_array), requires_grad=False
        )

    def log_pdf(self, distribution_params_input, log_prob_args):
        means = super().forward(distribution_params_input)
        normal = Normal(loc=means, scale=torch.ones_like(means) * self.stds)
        if self.is_truncated_to_output_bounds:
            left_bounds, right_bounds = (
                self.bounds_handler.bounds[:, 0],
                self.bounds_handler.bounds[:, 1],
            )
            return (
                normal.log_prob(log_prob_args)
                - torch.log(
                    normal.cdf(torch.ones_like(means) * right_bounds)
                    - normal.cdf(torch.ones_like(means) * left_bounds)
                )
            ).sum(axis=1)
        else:
            return normal.log_prob(log_prob_args).sum(axis=1)

    def forward(self, observation, is_means_only=False):
        mean = super().forward(observation)
        if is_means_only:
            return mean
        mean_numpy = mean.detach().cpu().numpy()
        if self.is_truncated_to_output_bounds:
            left_bounds, right_bounds = (
                self.output_bounds_array[None, :, 0],
                self.output_bounds_array[None, :, 1],
            )
            sampled_action = torch.FloatTensor(
                truncnorm(
                    loc=mean_numpy,
                    scale=np.ones_like(mean_numpy) * self.stds_array,
                    b=(right_bounds - mean_numpy) / self.stds_array,
                    a=(left_bounds - mean_numpy) / self.stds_array,
                )
                .rvs()
                .reshape(1, -1)
            )
        else:
            sampled_action = Normal(
                loc=mean, scale=self.stds * torch.ones_like(mean)
            ).sample()

        return sampled_action


class GaussianMeanStd(ModelNN):
    def __init__(
        self,
        mean: ModelNN,
        std: ModelNN,
        output_bounds: List[Union[List[float], np.ndarray]],
        is_truncated_to_output_bounds: bool = True,
    ):
        super().__init__()
        self.mean = mean
        self.std = std
        self.is_truncated_to_output_bounds = is_truncated_to_output_bounds
        self.bounds_handler = BoundsHandler(output_bounds)
        self.output_bounds_array = np.array(output_bounds)

        if is_truncated_to_output_bounds and output_bounds is None:
            raise ValueError(
                "output_bounds must be specified when is_truncated_to_output_bounds is True"
            )

    def forward(self, observation, is_means_only=False):
        mean, std = self.mean(observation), self.std(observation)
        # mean, std = self.mean(observation), self.std * torch.linalg.norm(observation)

        if is_means_only:
            return mean

        if self.is_truncated_to_output_bounds:
            mean_numpy = mean.detach().cpu().numpy()
            std_numpy = std.detach().cpu().numpy()
            left_bounds, right_bounds = (
                self.output_bounds_array[None, :, 0],
                self.output_bounds_array[None, :, 1],
            )
            sampled_action = torch.FloatTensor(
                truncnorm(
                    loc=mean_numpy,
                    scale=std_numpy,
                    b=(right_bounds - mean_numpy) / std_numpy,
                    a=(left_bounds - mean_numpy) / std_numpy,
                )
                .rvs()
                .reshape(1, -1)
            )
        else:
            sampled_action = Normal(loc=mean, scale=std).sample()

        return sampled_action

    def log_pdf(self, distribution_params_input, log_prob_args):
        means, stds = self.mean(distribution_params_input), self.std(
            distribution_params_input
        )
        normal = Normal(loc=means, scale=stds)
        if self.is_truncated_to_output_bounds:
            left_bounds, right_bounds = (
                self.bounds_handler.bounds[:, 0],
                self.bounds_handler.bounds[:, 1],
            )
            return (
                normal.log_prob(log_prob_args)
                - torch.log(
                    normal.cdf(torch.ones_like(means) * right_bounds)
                    - normal.cdf(torch.ones_like(means) * left_bounds)
                )
            ).sum(axis=1)
        else:
            return normal.log_prob(log_prob_args).sum(axis=1)

    def entropy(self, distribution_params_input):
        means, stds = self.mean(distribution_params_input), self.std(
            distribution_params_input
        )

        a = self.bounds_handler.bounds[:, 0] * torch.ones_like(means)
        b = self.bounds_handler.bounds[:, 1] * torch.ones_like(means)
        alpha = (a - means) / stds
        beta = (b - means) / stds

        standard_normal = Normal(
            loc=torch.zeros_like(means), scale=torch.ones_like(means)
        )
        z = standard_normal.cdf(beta) - standard_normal.cdf(alpha)

        phi_alpha = torch.exp(standard_normal.log_prob(alpha))
        phi_beta = torch.exp(standard_normal.log_prob(beta))

        return (
            torch.log(math.sqrt(2 * math.pi * math.e) * stds * z)
            + (alpha * phi_alpha - beta * phi_beta) / (2 * z)
        ).sum(axis=1)



"""Contains main `Optimizable` class.

`Optimizable` class is to be used normally as a parent class for all objects that need to be optimized.
"""

import numpy as np
from scipy.optimize import Bounds, NonlinearConstraint, minimize
from enum import Enum, auto

from regelum.utils import rg

try:
    from casadi import Opti

except (ModuleNotFoundError, ImportError):
    from unittest.mock import MagicMock

    Opti = MagicMock()


try:
    import torch

except ModuleNotFoundError:
    from unittest.mock import MagicMock

    torch = MagicMock()
    UpdatableSampler = MagicMock()

from typing import Callable, List, Tuple, Optional, Union, Dict

from .core.configs import OptimizerConfig
from .core.entities import (
    FunctionWithSignature,
    Hook,
    OptimizationVariable,
    VarContainer,
    FuncContainer,
    NestedFunction,
)
from .core.hooks import requires_grad, detach, data_closure

import regelum


class source_data_hook:
    def __init__(self, source, source_kwargs, func):
        self.source = source
        self.source_kwargs = source_kwargs
        self.func = func
        self.__name__ = "source_data_hook"

    def __call__(self, whatever):
        kwargs = {
            kwarg: var(True) if isinstance(var, OptimizationVariable) else var
            for kwarg, var in self.source_kwargs.items()
        }
        if self.source is None:
            return self.func(**kwargs)
        else:
            return self.func(self.source(True), **kwargs)


class source_metadata_hook:
    def __init__(self, source, source_kwargs, func):
        self.source = source
        self.source_kwargs = source_kwargs
        self.func = func
        self.__name__ = "source_metadata_hook"

    def __call__(self, whatever):
        kwargs = {
            kwarg: (
                var(with_metadata=True)
                if isinstance(var, OptimizationVariable)
                else var
            )
            for kwarg, var in self.source_kwargs.items()
        }
        if self.source is None:
            return self.func(**kwargs)
        else:
            return self.func(self.source(with_metadata=True), **kwargs)


class Optimizable(regelum.RegelumBase):
    """Base class for all optimizable objects.

    This class is to be used normally as a parent class for all objects that incapsulate optimization routines.
    However, you can also use it as a separate instance and use all methods outside of the `Optimizable` class.

    Args:
        optimizer_config (OptimizerConfig): An object of the
            OptimizerConfig class that defines the behavior of the
            optimizer.
    """

    def __init__(self, optimizer_config: OptimizerConfig) -> None:
        """Initialize an optimizable object.

        Args:
            optimizer_config (OptimizerConfig): An object of the
                OptimizerConfig class that defines the behavior of the
                optimizer.
        """
        self.optimizer_config = optimizer_config
        self.kind = optimizer_config.kind

        self.__variables: VarContainer = VarContainer([])
        self.__functions: FuncContainer = FuncContainer(tuple())
        self.params_changed = False
        self.is_check_status = True
        self.__is_problem_defined = False

        if self.kind == "symbolic":
            self.__opti_common = Opti()
            self.__opti = self.__opti_common.copy()
            self.__opt_func = None
            if optimizer_config.opt_method is None:
                optimizer_config.opt_method = "ipopt"
        elif self.kind == "numeric":
            if optimizer_config.opt_method is None:
                self.__bounds = None
                optimizer_config.opt_method = "SLSQP"
        elif self.kind == "tensor":
            self.optimizer = None
            if optimizer_config.opt_method is None:
                from torch.optim import Adam

                optimizer_config.opt_method = Adam
        else:
            raise NotImplementedError("Not implemented this kind of optimizer")

        self.opt_method = optimizer_config.opt_method
        self.__opt_options = optimizer_config.opt_options
        self.__log_options = optimizer_config.log_options
        self.opt_status = "unknown"

    @property
    def opt_options(self):
        """Get optimization options.

        Returns:
            dict: Optimization options
        """
        return self.__opt_options

    @property
    def log_options(self):
        """Get log options.

        Returns:
            dict: Log options
        """
        return self.__log_options

    @property
    def objective(self):
        """Get the objective.

        Returns:
            function/objective: Objective
        """
        assert len(self.objectives) == 1, "Ambiguous objective definition."
        return self.objectives[0]

    @property
    def objectives(self):
        """Get all objective functions.

        Returns:
            list: Objectives
        """
        return self.__functions.objectives

    @property
    def constraints(self):
        """Get all constraints.

        Returns:
            list: Constraints
        """
        return self.__functions.constraints

    @property
    def functions(self):
        """Get all functions.

        Returns:
            list: Functions
        """
        return self.__functions

    @property
    def variables(self) -> VarContainer:
        """Get all declared variables.

        Returns:
            VarContainer: Variables
        """
        return self.__variables

    def __recreate_opti(self):
        """Recreate optimization instance."""
        self.__opti = Opti()

    def __recreate_symbolic_variables(self):
        """Recreate symbolic variables in case if we use symbolic optimization and changed optimization procedure."""
        self.__recreate_opti()

        for variable in self.variables:
            dims = variable.dims
            if isinstance(dims, tuple):
                if len(dims) == 1:
                    dims = dims[0]

            is_constant = variable.is_constant
            if isinstance(dims, tuple):
                metadata = (
                    self.__opti.variable(*dims)
                    if not is_constant
                    else self.__opti.parameter(*dims)
                )
            elif isinstance(dims, int):
                metadata = (
                    self.__opti.variable(dims)
                    if not is_constant
                    else self.__opti.parameter(dims)
                )
            else:
                raise ValueError("Unknown dimensions format")

            variable.with_metadata(metadata, inplace=True)

        self.__recreate_symbolic_functions()

    def __recreate_symbolic_functions(self):
        """Recreate symbolic functions in case if we use symbolic optimization and changed optimization procedure."""
        __functions = sum(
            [
                function
                for function in self.functions
                if not (
                    ("__bound" in function.name) and (function.variables[0].is_constant)
                )
            ]
        )
        for function in __functions:
            function: FunctionWithSignature
            for variable in function.variables:
                if variable.name == "var":
                    variable.metadata = variable.data.metadata

            function = self.__infer_and_register_symbolic_prototype(
                function, function.variables
            )
            metafunc = function.metadata
            if function.is_objective:
                self.__opti.minimize(metafunc)
            else:
                self.__opti.subject_to(metafunc <= 0)

    def __refresh_binded_variables(self):
        """Refresh all bound variables in the function list by updating their status."""
        for function in self.functions:
            for variable in function.variables:
                if isinstance(variable.data, OptimizationVariable):
                    variable.is_constant = variable.data.is_constant

    def __fix_variables_tensor(self, variables_to_fix, data_dict, metadata_dict):
        """Fix variables in case of self.kind == 'tensor' by setting them as constants.

        Args:
            variables_to_fix (List[str]): List of variable names to be
                fixed.
            data_dict (dict): Dictionary mapping variable names to their
                corresponding data.
            metadata_dict (dict): Dictionary mapping variable names to
                their corresponding metadata.
        """
        self.optimizer = None
        self.__variables.fix(variables_to_fix, hook=Hook(detach, act_on="data"))

    def __fix_variables_symbolic(self, variables_to_fix, data_dict, metadata_dict):
        """Fix variables in case of self.kind == 'symbolic' by setting them as constants.

        Args:
            variables_to_fix (List[str]): List of variable names to be
                fixed.
            data_dict (dict): Dictionary mapping variable names to their
                corresponding data.
            metadata_dict (dict): Dictionary mapping variable names to
                their corresponding metadata.
        """
        if metadata_dict is None:
            metadata_dict = {}
        passed_unfixed_variables = self.variables.selected(variables_to_fix)
        assert isinstance(
            passed_unfixed_variables, VarContainer
        ), "An error occured while fixing variables."
        self.__variables.fix(variables_to_fix)
        self.__refresh_binded_variables()
        self.__recreate_symbolic_variables()

        self.params_changed = True

    def fix_variables(
        self,
        variables_to_fix: List[str],
        data_dict: Optional[Dict] = None,
        metadata_dict: Optional[Dict] = None,
    ):
        """Fix variables by setting them as constants.

        Args:
            variables_to_fix (List[str]): List of variable names to be
                fixed.
            data_dict (dict, optional): Optional dictionary mapping
                variable names to their corresponding data, defaults to
                None.
            metadata_dict (dict, optional): Optional dictionary mapping
                variable names to their corresponding metadata, defaults
                to None.
        """
        if self.kind == "tensor":
            self.__fix_variables_tensor(
                variables_to_fix, data_dict=data_dict, metadata_dict=metadata_dict
            )

        elif self.kind == "symbolic":
            self.__fix_variables_symbolic(
                variables_to_fix, data_dict=data_dict, metadata_dict=metadata_dict
            )
        else:
            self.__variables.fix(variables_to_fix)

    def __unfix_variables_tensor(self, variables_to_unfix):
        """Unfix tensor variables by setting them as non-constants.

        Args:
            variables_to_unfix (List[str]): List of variable names to be
                unfixed.
        """
        self.optimizer = None
        self.__variables.unfix(
            variables_to_unfix, hook=Hook(requires_grad, act_on="data")
        )

    def __unfix_variables_symbolic(self, variables_to_unfix):
        """Unfix symbolic variables by setting them as non-constants.

        Args:
            variables_to_unfix (List[str]): List of variable names to be
                unfixed.
        """
        passed_fixed_variables = self.variables.selected(variables_to_unfix)
        assert isinstance(
            passed_fixed_variables, VarContainer
        ), "An error occured while fixing variables."

        self.__variables.unfix(variables_to_unfix)
        self.__refresh_binded_variables()
        self.__recreate_symbolic_variables()

        self.params_changed = True

    def unfix_variables(
        self,
        variables_to_unfix: List[str],
    ):
        """Unfix variables by setting them as non-constants.

        Args:
            variables_to_unfix (List[str]): List of variable names to be
                unfixed.
        """
        if self.kind == "tensor":
            self.__unfix_variables_tensor(variables_to_unfix=variables_to_unfix)
        elif self.kind == "symbolic":
            self.__unfix_variables_symbolic(variables_to_unfix=variables_to_unfix)
        else:
            self.__variables.unfix(variables_to_unfix)

    def create_variable_metadata(self, *dims, is_constant=False, like=None):
        """Create metadata for a variable based on dimensions and characteristics.

        This method handles the creation of metadata for a variable which can either be
        a symbolic variable or a parameter, depending on the `is_constant` flag and
        whether the `like` parameter is used to clone an existing variable's properties.

        Args:
            *dims: Variable dimensions, can be a series of integers or
                tuples.
            is_constant: Flag indicating if the variable should be
                treated as a constant. Defaults to False.
            like: An existing variable whose shape and metadata should
                be replicated. Must have a `shape` attribute if
                provided.

        Returns:
            The created metadata object for the variable.

        Raises:
            AssertionError: If `like` doesn't have a `shape` attribute,
                or if the provided dimensions are invalid for symbolic
                variables.
        """
        metadata = None
        if self.kind == "symbolic":
            if like is not None:
                assert hasattr(
                    like, "shape"
                ), "Symbolic variable prototype must have shape"
                dims = like.shape
            if len(dims) == 1:
                assert isinstance(
                    dims[0], (int, tuple)
                ), "Dimension must be integer or tuple"
                if isinstance(dims[0], tuple):
                    assert len(dims[0]) <= 2, "Symbolic variable dimension must be <= 2"
                    metadata = (
                        self.__opti.variable(*dims[0])
                        if not is_constant
                        else self.__opti.parameter(*dims[0])
                    )
                else:
                    metadata = (
                        self.__opti.variable(dims[0])
                        if not is_constant
                        else self.__opti.parameter(dims[0])
                    )
            else:
                metadata = (
                    self.__opti.variable(*dims)
                    if not is_constant
                    else self.__opti.parameter(*dims)
                )
        elif self.kind == "tensor":
            metadata = like
        return metadata

    def create_variable(
        self,
        *dims,
        name: str,
        is_constant=False,
        like=None,
        is_nested_function=False,
        nested_variables=None,
    ):
        """Create a new optimization variable or a nested function.

        Based on the provided dimensions and characteristics, this method either creates
        a standard optimization variable or a nested function variable. The newly created
        variable is then added to the internal variables container.

        Args:
            *dims: Dimensions of the variable.
            name: The name of the variable.
            is_constant: Specifies if the variable is a constant.
                Defaults to False.
            like: Optionally, an existing variable to base the new
                variable on.
            is_nested_function: Flag to determine if the variable is a
                nested function. Defaults to False.
            nested_variables: A list of variables that are nested within
                this variable, only used if `is_nested_function` is
                True.

        Returns:
            The newly created variable or nested function.
        """
        metadata = None
        nested_variables = [] if nested_variables is None else nested_variables
        if not is_nested_function:
            metadata = self.create_variable_metadata(
                *dims, is_constant=is_constant, like=like
            )
            new_variable = OptimizationVariable(
                name=name,
                dims=dims,
                metadata=metadata,
                data=like,
                is_constant=is_constant,
            )
        else:
            new_variable = NestedFunction(
                name=name,
                dims=dims,
                data=like,
                metadata=metadata,
                is_constant=is_constant,
                nested_variables=VarContainer(nested_variables),
            )
        if self.kind == "tensor" or self.kind == "numeric":
            if like is not None:
                new_variable = new_variable.with_data(like).with_metadata(like)
        self.__variables = self.__variables + new_variable

        return new_variable

    def __infer_and_register_symbolic_prototype(
        self, func: FunctionWithSignature, variables: VarContainer
    ):
        """Infer and register a symbolic function prototype using provided variables.

        This internal method uses the provided function signature and variable container
        to infer metadata and register a symbolic prototype of the function.

        Args:
            func: The function with a signature to infer metadata for.
            variables: A container of variables to use for the function.

        Returns:
            The function with inferred metadata.
        """
        metadata = func(
            **variables.to_metadata_dict(), with_metadata=True, raw_eval=True
        )
        func.metadata = metadata
        return func

    def register_objective(
        self,
        func: Callable,
        variables: Union[
            List[OptimizationVariable], VarContainer, OptimizationVariable
        ],
    ):
        """Register an objective function for optimization.

        The given function is wrapped with a function signature and is registered as an
        objective function within the internal functions container. It also ensures that
        the variables are properly contained within a VarContainer.

        Args:
            func: The objective function to be registered.
            variables: The variables over which the objective function
                is defined. Can be a single variable, a list of
                variables, or a VarContainer.

        Raises:
            AssertionError: If the new container for functions is None
                after attempting to register the objective.
        """
        func = FunctionWithSignature(func, is_objective=True)
        if isinstance(variables, OptimizationVariable):
            variables = [variables]
        if isinstance(variables, List):
            variables = VarContainer(variables)
        func.declare_variables(variables)

        if self.kind == "symbolic":
            self.__register_symbolic_objective(
                func,
                variables=variables,
            )
        else:
            pass

        new_container = self.__functions + func
        assert new_container is not None, f"Couldn't register objective {func}"
        self.__functions = new_container

    @staticmethod
    def connect_source(
        connect_to: OptimizationVariable,
        func: Callable,
        source: Optional[OptimizationVariable] = None,
        act_on="all",
        discard_prev_sources=True,
        **source_kwargs,
    ):
        """Connect a source variable to another optimization variable, applying a given function.

        This method establishes a connection between a source optimization variable and the target
        optimization variable such that the specified function is applied to the source before
        being passed to the target. The function can take additional keyword arguments from
        `source_kwargs`.

        Args:
            connect_to: The optimization variable to which the source
                should be connected.
            func: The function to apply to the source variable before
                passing its value to `connect_to`.
            source: The source optimization variable, if any. If None,
                only `source_kwargs` are used.
            act_on: Specifies whether to act on "data", "metadata", or
                "all". Defaults to "all".
            discard_prev_sources: If True, previous sources connected to
                `connect_to` will be discarded. Defaults to True.
            **source_kwargs: Additional keyword arguments to pass to the
                function alongside the source variable.
        """
        data_hook = Hook(source_data_hook(source, source_kwargs, func), act_on="data")
        metadata_hook = Hook(
            source_metadata_hook(source, source_kwargs, func), act_on="metadata"
        )

        if discard_prev_sources:
            for hook in connect_to.hooks:
                if "source" in hook.name:
                    connect_to.discard_hook(hook, disable_only=False)

        if act_on in ["all", "data"]:
            connect_to.register_hook(data_hook, first=True)
        if act_on in ["all", "metadata"]:
            connect_to.register_hook(metadata_hook, first=True)

    def __register_symbolic_objective(
        self,
        func: FunctionWithSignature,
        variables: VarContainer,
    ):
        """Register a symbolic objective function into the optimization process.

        This internal method wraps the given function with additional metadata and sets it as the objective
        to minimize in the symbolic optimizer.

        Args:
            func: The objective function to be registered and minimized.
            variables: The container holding variables over which the
                objective function is defined.
        """
        func = self.__infer_and_register_symbolic_prototype(func, variables)
        self.__opti.minimize(func.metadata)

    @staticmethod
    def handle_bounds(
        bounds: Union[list, np.ndarray, None],
        dim_variable: int,
        tile_parameter: int = 0,
    ) -> Tuple:
        """Process bounds for optimization variables and prepares them for use in optimization.

        This static method handles the bounds for each dimension of an optimization variable,
        returning a tuple with the prepared bounds arrays necessary for the optimization process.

        Args:
            bounds: The bounds for each dimension of the variable.
                Defaults to (-inf, inf) if None.
            dim_variable: The number of dimensions of the optimization
                variable.
            tile_parameter: The number of times to tile the variable.
                Defaults to 0, meaning no tiling.

        Returns:
            A tuple containing the processed bounds, initial guess, and
            min-max ranges for the optimization variable.
        """
        if bounds is None:
            assert dim_variable is not None, "Dimension of the bounds must be specified"
            bounds = np.array([[-np.inf, np.inf] for _ in range(dim_variable)])
        else:
            assert isinstance(
                bounds, (list, np.ndarray)
            ), "bounds must be a list or ndarray"
            if not isinstance(bounds, np.ndarray):
                bounds = np.array(bounds)
            assert len(bounds.shape) == 2, (
                f"bounds must be of shape ({dim_variable}, 2)."
                + f" You have ({bounds.shape[0]}, {bounds.shape[1]}"
            )
            assert bounds.shape[0] == dim_variable, (
                f"bounds should be of size ({dim_variable}, 2)."
                + f" You have ({bounds.shape[0]}, {bounds.shape[1]})."
            )

        for i, row in enumerate(bounds):
            if row[0] > row[1]:
                raise ValueError(
                    "The lower bound of action is greater"
                    + " than the upper bound of action "
                    + f"at index {i} ({row[0] > row[1]})"
                )
        variable_min = rg.force_row(bounds[:, 0])
        variable_max = rg.force_row(bounds[:, 1])
        variable_initial_guess = (variable_min + variable_max) / 2
        if tile_parameter > 0:
            variable_sequence_initial_guess = rg.rep_mat(
                variable_initial_guess, 1, tile_parameter
            )
            sequence_min = rg.rep_mat(variable_min, tile_parameter, 1)
            sequence_max = rg.rep_mat(variable_max, tile_parameter, 1)
            result_bounds = rg.hstack((sequence_min, sequence_max))
            variable_initial_guess = variable_sequence_initial_guess
        else:
            result_bounds = bounds
        return result_bounds, variable_initial_guess, variable_min, variable_max

    def register_bounds(
        self, variable_to_bound: OptimizationVariable, bounds: np.ndarray
    ):
        assert isinstance(
            variable_to_bound, OptimizationVariable
        ), "variable_to_bound should be of type OptimizationVariable, "
        f"not {type(variable_to_bound)}"

        if self.kind == "symbolic":
            self.__register_symbolic_bounds(variable_to_bound, bounds)

        elif self.kind == "numeric":
            self.__register_numeric_bounds(bounds)

        elif self.kind == "tensor":
            self.__register_tensor_bounds(bounds)

    def __register_symbolic_bounds(
        self, variable_to_bound: OptimizationVariable, bounds: np.ndarray
    ):
        self.__bounds = bounds

        def lb_constr(var):
            return bounds[:, : var.shape[1]] - var

        def ub_constr(var):
            return var - bounds[:, var.shape[1] :]

        var = variable_to_bound.renamed("var", inplace=False)
        var.data = variable_to_bound
        self.register_constraint(
            lb_constr, variables=[var], name=f"{variable_to_bound.name}__bound_lower"
        )
        self.register_constraint(
            ub_constr, variables=[var], name=f"{variable_to_bound.name}__bound_upper"
        )

    def __register_numeric_bounds(self, bounds):
        self.__bounds = Bounds(
            bounds[:, 0],
            bounds[:, 1],
            keep_feasible=True,
        )

    def __register_tensor_bounds(self, bounds):
        self.__bounds = bounds

    def register_constraint(
        self, func: Callable, variables: List[OptimizationVariable], name=None
    ):
        func = FunctionWithSignature(func)
        if name is not None:
            func.name = name

        _variables = VarContainer(variables)
        func.declare_variables(_variables)
        if self.kind == "symbolic":
            self.__register_symbolic_constraint(func, _variables)
        elif self.kind == "numeric":
            self.__register_numeric_constraint(func, _variables)
        elif self.kind == "tensor":
            self.__register_tensor_constraint(func, _variables)

        new_container = self.__functions + func
        assert new_container is not None, f"Couldn't register objective {func}"

        self.__functions = new_container

    def __register_symbolic_constraint(
        self, func: FunctionWithSignature, variables: VarContainer
    ):
        """Register a symbolic constraint into the optimization problem.

        This method infers the metadata for the provided functional constraint and formulates
        it as a symbolic constraint that the optimizer is subject to.

        Args:
            func: The constraint function with a signature.
            variables: The container of variables that the function
                operates on.
        """
        func = self.__infer_and_register_symbolic_prototype(func, variables)
        constr = rg.vec(func.metadata) <= -1e-8
        self.__opti.subject_to(constr)

    def __register_numeric_constraint(
        self, func: FunctionWithSignature, variables: VarContainer
    ):
        """Register a numeric constraint into the optimization problem.

        Currently, this method serves as a placeholder for future implementations where
        numerical constraints would be processed differently from symbolic ones.

        Args:
            func: The constraint function with a signature.
            variables: The container of variables that the function
                operates on.
        """
        func.declare_variables(variables)

    def __register_tensor_constraint(
        self, func: FunctionWithSignature, variables: VarContainer
    ):
        """Register a tensor constraint into the optimization problem.

        Currently, this method serves as a placeholder for future implementations where
        tensor constraints would be processed differently from symbolic ones.

        Args:
            func: The constraint function with a signature.
            variables: The container of variables that the function
                operates on.
        """
        func.declare_variables(variables)

    @property
    def constants(self):
        """Return the constants from the variables container.

        Returns:
            An object representing the constants in the optimization
            problem.
        """
        return self.__variables.constants

    @property
    def decision_variables(self):
        """Return the decision variables from the variables container.

        Returns:
            An object representing the decision variables in the
            optimization problem.
        """
        return self.__variables.decision_variables

    def substitute_parameters(self, **parameters):
        """Substitute parameters in the optimization problem with given values.

        This method is responsible for substituting the values of parameters (constants or
        initial guesses for decision variables) before optimization is carried out.

        Args:
            **parameters: Keyword arguments representing the parameter
                names and their values to substitute.
        """
        self.variables.substitute_data(**parameters)
        if self.kind == "symbolic":
            params_to_delete = []
            for k, v in parameters.items():
                if k in self.variables:
                    if self.variables[k].is_constant:
                        self.__opti.set_value(self.variables[k](with_metadata=True), v)
                    else:
                        self.__opti.set_initial(
                            self.variables[k](with_metadata=True), v
                        )
                        params_to_delete.append(k)
            for k in params_to_delete:
                del parameters[k]
        else:
            self.variables.substitute_metadata(**parameters)

    def is_target_event(self, event):
        """Check if an event is one of the target events for callback.

        Args:
            event: The event to check.

        Returns:
            True if the event is a target event, False otherwise.
        """
        if self.__callback_target_events is None:
            return False

        return event in self.__callback_target_events

    def optimize_on_event(self, **parameters):
        """Optimize the problem based on an event.

        This method is not implemented and serves as a placeholder for future functionality
        where optimization would be triggered by specific events.

        Raises:
            NotImplementedError: Always, as this method is not
                implemented.
        """
        raise NotImplementedError("optimize_on_event is not implemented")

    def optimize(self, raw=False, is_constrained=True, **parameters):
        """Optimize the problem based on the kind of optimizer.

        This method delegates the optimization to the appropriate method based on the optimizer's kind,
        whether it's symbolic (CasADi), numeric (SciPy), or tensor (Torch)-based.

        Args:
            raw: Determines if the raw optimization results should be
                returned. Defaults to False.
            is_constrained: Flags if the optimization includes
                constraints. Defaults to True.
            **parameters: Additional parameters for the optimization
                process.

        Returns:
            The result of the optimization process.

        Raises:
            NotImplementedError: If the optimizer kind is not
                recognized.
        """
        if not self.__is_problem_defined:
            self.define_problem()

        result = None
        if self.kind == "symbolic":
            result = self.optimize_symbolic(**parameters, raw=raw)
        elif self.kind == "numeric":
            result = self.optimize_numeric(**parameters, raw=raw)
        elif self.kind == "tensor":
            self.optimize_tensor(**parameters, is_constrained=is_constrained)
            result = self.variables.decision_variables
        else:
            raise NotImplementedError

        return result

    @property
    def opt_func(self):
        """Return the optimization function.

        Returns:
            The function used for optimization.
        """
        return self.__opt_func

    @property
    def opti(self):
        """A property that returns the current Opti instance.

        Returns:
            The Opti instance used for managing the optimization
            problem.
        """
        return self.__opti

    def optimize_symbolic(self, raw=True, tol=1e-12, **kwargs):
        """Optimize the symbolic (CasADi) problem with respect to the objective and constraints.

        This method prepares and executes the optimization process for a symbolic optimization
        problem. It generates a callable optimization function if it's not already defined or
        if the parameters have changed since the last optimization.

        Args:
            raw: If True, returns the raw optimization result as is. If
                False, filters the result to only include decision
                variables' names and their values. Defaults to True.
            tol: The tolerance for considering the optimization
                successful. Defaults to 1e-12.
            **kwargs: Additional keyword arguments used for parameter
                substitution before optimization.

        Returns:
            The optimization result. The structure depends on the `raw`
            parameter.
        """
        if self.__opt_func is None or self.params_changed:
            self.__opti.solver(
                self.opt_method, dict(self.__log_options), dict(self.__opt_options)
            )
            self.var_for_opti = VarContainer(
                [
                    variable
                    for variable in self.variables
                    if all(
                        [
                            "source" not in hook_name
                            for hook_name in variable.hooks.hooks_container.names
                        ]
                    )
                ]
            )
            self.__opt_func = self.__opti.to_function(
                "min_fun",
                [
                    variable(with_metadata=True)
                    for variable in self.var_for_opti.constants
                ],
                [
                    *[
                        variable(with_metadata=True)
                        for variable in self.var_for_opti.decision_variables
                    ],
                    self.objective.metadata,
                    *[func.metadata for func in self.constraints],
                ],
                list(self.var_for_opti.constants.names),
                [
                    *self.var_for_opti.decision_variables.names,
                    self.objective.name,
                    *[func.name for func in self.constraints],
                ],
            )
            self.params_changed = False

        self.substitute_parameters(**kwargs)

        result = self.__opt_func(
            **{k: v for k, v in self.var_for_opti.constants.to_data_dict().items()}
        )
        if self.is_check_status:
            self.update_status(result, tol=tol)
        return (
            result
            if raw
            else {
                name: value
                for name, value in result.items()
                if name in self.decision_variables.names
            }
        )

    def update_status(self, result=None, tol=1e-8):
        """Update the optimization status based on the result.

        This method evaluates the result against a tolerance level to set the optimization status.
        If any constraint exceeds the tolerance, the status is set to failed.

        Args:
            result: The result of the optimization process.
            tol: The tolerance level used to evaluate constraints
                satisfaction. Defaults to 1e-8.
        """
        self.opt_status = OptStatus.success
        if self.kind == "symbolic":
            for constr_name in self.constraints.names:
                if rg.max(result[constr_name]) > tol:
                    self.opt_status = OptStatus.failed
                    break

    def optimize_numeric(self, raw=False, **parameters):
        """Optimize the numeric problem by using SciPy's minimize function.

        This method sets up and solves a numeric optimization problem. It substitutes parameters,
        defines constraints, and executes the optimization using the specified numerical method.

        Args:
            raw: If True, returns the raw `OptimizeResult` object from
                SciPy's minimize function. If False, returns the
                optimized decision variables as a NumPy array. Defaults
                to False.
            **parameters: Keyword arguments that include parameters to
                be substituted and any additional arguments for the
                optimization algorithm.

        Returns:
            The result of the numeric optimization. The structure
            depends on the `raw` parameter.

        Raises:
            AssertionError: If the initial guess is not provided and
                cannot be inferred from the decision variables.
        """
        self.substitute_parameters(**parameters)
        constraints = [
            NonlinearConstraint(func, -np.inf, 0) for func in self.constraints
        ]
        initial_guess = parameters.get(self.decision_variables.names[0])
        if initial_guess is None:
            assert (
                self.decision_variables[0].data is not None
            ), "Initial guess is None. Try to set 'like=...' when creating a decision variable"

            initial_guess = self.decision_variables[0].data

        opt_result = minimize(
            self.objective,
            x0=initial_guess,
            method=self.opt_method,
            bounds=self.__bounds if hasattr(self, "__bounds") else None,
            options=self.__opt_options,
            constraints=constraints,
            tol=1e-7,
        )
        return opt_result if raw else opt_result.x

    def instantiate_configuration(self):
        """Instantiate the optimizer configuration for the optimization process.

        This method prepares the optimizer based on the decision variables and the optimization method specified.
        It also validates the configuration and ensures that a single decision variable and a single objective are present
        (In the current verion we consider optimization problem ambiguous if there are multiple decision variables or multiple objectives.
        If you want to optimize on several variables alternately, use `fix_variables` and `unfix_variables` explicitly).

        Returns:
            A tuple containing the number of epochs to run and the
            objective function.

        Raises:
            AssertionError: If the optimization method is not callable,
                if there are multiple or no decision variables, or if
                multiple objectives are present.
        """
        options = self.optimizer_config.config_options
        if self.optimizer is None:
            assert self.opt_method is not None and callable(
                self.opt_method
            ), f"Wrong optimization method {self.opt_method}."
            vars_to_opt = VarContainer(
                [
                    variable
                    for variable in self.decision_variables
                    if all(
                        [
                            "source" not in hook_name
                            for hook_name in variable.hooks.hooks_container.names
                        ]
                    )
                ]
            )
            assert len(vars_to_opt) == 1, "ambigous optimization variables"
            dvar = vars_to_opt[0]
            assert dvar is not None, "Couldn't find decision variable"
            assert isinstance(dvar, OptimizationVariable), "Something went wrong..."
            self.optimizer = self.opt_method(
                data_closure(dvar(with_metadata=False))(True), **self.__opt_options
            )
        n_epochs = options.get("n_epochs") if options.get("n_epochs") is not None else 1
        assert isinstance(n_epochs, int), "n_epochs must be an integer"
        assert len(self.functions.objectives) == 1, "Only one objective is supported"
        objective = self.functions.objectives[0]
        assert isinstance(objective, FunctionWithSignature), "Something went wrong..."

        return n_epochs, objective

    def eval_contraints(self):
        """Evaluate the constraints of the optimization problem.

        This method applies a penalty function to each constraint and sums up their values.

        Returns:
            The sum of the penalty functions applied to the constraints.
        """
        return sum(
            [
                rg.penalty_function(
                    func(**self.variables.to_data_dict()),
                    penalty_coeff=5,
                    delta=-3,
                )
                for func in self.constraints
            ]
        )

    def opt_constraints_tensor(self):
        """Perform constrained optimization for a tensor (Torch)-based optimizer.

        This method iterates over the constraints and performs gradient descent steps
        to minimize the constraints violations.

        Sets the optimization status to failed if any constraint is violated after optimization.
        """
        n_epochs_per_constraint = 1
        if self.optimizer_config.config_options["constrained_optimization_policy"][
            "is_activated"
        ]:
            n_epochs_per_constraint = self.optimizer_config.config_options[
                "constrained_optimization_policy"
            ]["defaults"]["n_epochs_per_constraint"]
        for _ in range(n_epochs_per_constraint):
            self.optimizer.zero_grad()
            constr_value = self.eval_contraints()
            if (
                max(
                    [func(**self.variables.to_data_dict()) for func in self.constraints]
                ).item()
                < 0
            ):
                break
            constr_value.backward()
            self.optimizer.step()

        if (
            max(
                [func(**self.variables.to_data_dict()) for func in self.constraints]
            ).item()
            > 0
        ):
            self.opt_status = OptStatus.failed
        else:
            self.opt_status = OptStatus.success

        print(constr_value)

    @apply_callbacks()
    def post_epoch(self, epoch_idx: int, objective_epoch_history: List[float]):
        """Emit post-epoch event.

        This a callback method called after each epoch during tensor-based optimization and is used to perform custom actions after each epoch, such as logging or saving epoch history.

        Args:
            epoch_idx: The index of the current epoch.
            objective_epoch_history: A list of objective function values
                from the current epoch.

        Returns:
            A tuple containing the current epoch index and the objective
            epoch history.
        """
        return epoch_idx, objective_epoch_history

    @apply_callbacks()
    def post_optimize(self):
        """Emit post-optimize event.

        This a callback method called after the completion of the optimization process and can be used to perform cleanup or finalization steps once optimization is finished.
        """
        pass

    def optimize_tensor_batch_sampler(
        self, batch_sampler, n_epochs, objective, is_constrained
    ):
        """Optimize the problem using a tensor (Torch)-based optimizer with batch sampling.

        This method iterates over the provided batch samples to perform epochs of optimization,
        optionally including constraint handling if specified.

        Args:
            batch_sampler: An iterable that provides batch samples for
                optimization.
            n_epochs: The number of epochs to run the optimization.
            objective: The objective function to be minimized.
            is_constrained: A boolean indicating whether to enforce
                constraints during optimization.
        """
        for epoch_idx in range(n_epochs):
            objective_epoch_history = []
            for batch_sample in batch_sampler:
                self.optimizer.zero_grad()
                self.substitute_parameters(**batch_sample)
                if is_constrained and len(self.constraints) > 0:
                    self.opt_constraints_tensor()
                    objective_value = objective(**self.variables.to_data_dict())
                    (objective_value + self.eval_contraints()).backward()
                else:
                    objective_value = objective(**self.variables.to_data_dict())
                    objective_value.backward()
                self.optimizer.step()
                objective_epoch_history.append(objective_value.item())
                self.post_epoch(epoch_idx, objective_epoch_history)

        if (
            self.optimizer_config.config_options.get("is_reinstantiate_optimizer")
            is not None
        ):
            # Force to recreate torch.optim.Optimizer object before every optimization
            if self.optimizer_config.config_options.get("is_reinstantiate_optimizer"):
                self.optimizer = None

    def optimize_tensor(self, **parameters):
        n_epochs, objective = self.instantiate_configuration()
        batch_sampler = parameters.get("batch_sampler")
        is_constrained = parameters.get("is_constrained")
        if batch_sampler is not None:
            self.optimize_tensor_batch_sampler(
                batch_sampler=batch_sampler,
                n_epochs=n_epochs,
                objective=objective,
                is_constrained=is_constrained,
            )
        else:
            self.substitute_parameters(**parameters)
            for epoch_idx in range(n_epochs):
                self.optimizer.zero_grad()
                if is_constrained and len(self.constraints) > 0:
                    self.opt_constraints_tensor()
                    objective_value = objective(**self.variables.to_data_dict())
                    objective_value_constrained = +self.eval_contraints()
                    objective_value_constrained.backward()
                else:
                    objective_value = objective(**self.variables.to_data_dict())
                    objective_value.backward()
                self.optimizer.step()
                self.post_epoch(epoch_idx, [objective_value.item()])
            if self.optimizer_config.config_options.get("is_reinstantiate_optimizer"):
                self.optimizer = None

        self.post_optimize()

    def define_problem(self):
        """Implement this method if you want to separate an optimization problem definition
        from the rest of your code. Needs only for readability.
        """
        self.__is_problem_defined = True

    def get_data_buffer_batch_size(self):
        """Retrieve the batch size for data sampling from the optimizer configuration.

        The method looks up the configuration for the data buffer sampling method and its associated
        keyword arguments to determine the batch size for sampling.

        Returns:
            The batch size determined by the data buffer sampling
            method.

        Raises:
            AssertionError: If the data_buffer_sampling_method or
                data_buffer_sampling_kwargs is not specified in the
                optimizer configuration.
            ValueError: If an unknown data_buffer_sampling_method is
                specified.
        """
        config_options = self.optimizer_config.config_options

        method_name = config_options.get("data_buffer_sampling_method")
        kwargs = config_options.get("data_buffer_sampling_kwargs")
        assert (
            method_name is not None
        ), "Specify `data_buffer_sampling_method` in your optimizer_config"
        assert (
            kwargs is not None
        ), "Specify `data_buffer_sampling_kwargs` in your optimizer_config"

        if method_name == "iter_batches":
            return kwargs.get("batch_size")
        elif method_name == "sample_last":
            return kwargs.get("n_samples") if kwargs.get("n_samples") is not None else 1
        else:
            raise ValueError("Unknown data_buffer_sampling_method")


class OptStatus(Enum):
    """Enum class representing optimization status."""

    success = auto()
    failed = auto()



"""Module that contains state or observation (depending on the context) predictors."""

from abc import ABC, abstractmethod

import regelum
from .utils import rg
from .system import System
from .model import ModelWeightContainer, ModelWeightContainerTorch
import torch


class Predictor(regelum.RegelumBase, ABC):
    """Blueprint of a predictor."""

    def __init__(
        self,
        system: System,
        pred_step_size: float,
    ):
        """Initialize an instance of a predictor.

        Args:
            system (System): System of which states are predicted
            pred_step_size (float): time interval between successive
                predictions
        """
        self.system = system
        self.pred_step_size = pred_step_size

    @abstractmethod
    def predict(self):
        pass

    def predict_state_sequence_from_action_sequence(
        self,
        state,
        action_sequence,
        is_predict_last: bool,
        return_predicted_states_only=False,
    ):
        """Predicts the state sequence from the given action sequence.

        Args:
            state (numpy.ndarray): The initial state.
            action_sequence (numpy.ndarray): The sequence of actions.
            is_predict_last (bool): Whether to predict the last state or
                not.
            return_predicted_states_only (bool, optional): Whether to
                return only the predicted states or both the predicted
                states and the action sequence. Defaults to False.

        Returns:
            tuple or numpy.ndarray: If `return_predicted_states_only` is
            False, returns a tuple containing the predicted state
            sequence and the action sequence. If
            `return_predicted_states_only` is True, returns only the
            predicted state sequence.
        """
        len_state_sequence = action_sequence.shape[0] - int(not is_predict_last)
        predicted_state_sequence = rg.zeros(
            [len_state_sequence, self.system.dim_state],
            prototype=action_sequence,
        )
        current_state = state
        for k in range(len_state_sequence):
            current_action = action_sequence[k, :]
            next_state = self.predict(current_state, current_action)
            predicted_state_sequence[k, :] = next_state
            current_state = next_state

        return (
            (predicted_state_sequence, action_sequence)
            if not return_predicted_states_only
            else predicted_state_sequence
        )

    def predict_state_sequence_from_model(
        self,
        state,
        prediction_horizon,
        is_predict_last: bool,
        model,
        model_weights=None,
        return_predicted_states_only=False,
    ):
        """Predicts a sequence of states from a given initial state using a model.

        Args:
            state: The initial state.
            prediction_horizon: The number of steps to predict into the
                future.
            is_predict_last: A boolean indicating whether to predict the
                last state.
            model: The model used for prediction.
            model_weights: The weights of the model (optional).
            return_predicted_states_only: A boolean indicating whether
                to return only the predicted states.

        Returns:
            If `return_predicted_states_only` is `True`, returns the
            predicted state sequence. Otherwise, returns a tuple
            containing the predicted state sequence and the action
            sequence.
        """
        if isinstance(model, ModelWeightContainer):
            if model_weights is not None:
                assert model_weights.shape[0] == prediction_horizon + 1 - int(
                    is_predict_last
                ), f"model_weights.shape[0] = {model_weights.shape[0]} should have length prediction_horizon + 1 - int(is_predict_last) = {prediction_horizon + 1 - int(is_predict_last)}"
                return self.predict_state_sequence_from_action_sequence(
                    state,
                    action_sequence=model_weights,
                    is_predict_last=is_predict_last,
                    return_predicted_states_only=return_predicted_states_only,
                )
            else:
                assert model._weights.shape[0] == prediction_horizon + 1 - int(
                    is_predict_last
                ), "model._weights.shape[0] should have length prediction_horizon + 1 - int(is_predict_last)"
                return self.predict_state_sequence_from_action_sequence(
                    state,
                    action_sequence=model._weights,
                    is_predict_last=is_predict_last,
                    return_predicted_states_only=return_predicted_states_only,
                )
        elif isinstance(model, ModelWeightContainerTorch):
            assert model._weights.shape[0] == prediction_horizon + 1 - int(
                is_predict_last
            ), "model._weights.shape[0] should have length prediction_horizon + 1 - int(is_predict_last)"
            dummy_input = torch.zeros(
                [
                    prediction_horizon + 1 - int(is_predict_last),
                    self.system.dim_observation,
                ],
            )
            return self.predict_state_sequence_from_action_sequence(
                state,
                action_sequence=model(dummy_input),
                is_predict_last=is_predict_last,
                return_predicted_states_only=return_predicted_states_only,
            )

        predicted_state_sequence = rg.zeros(
            [prediction_horizon + 1 - int(not is_predict_last), self.system.dim_state],
            prototype=state,
        )
        action_sequence = rg.zeros(
            [prediction_horizon + 1, self.system.dim_state],
            prototype=state,
        )
        current_state = state
        for k in range(prediction_horizon + 1):
            action_sequence[k, :] = model(current_state, model_weights)
            next_state = self.predict(current_state, action_sequence[k, :])
            if k < predicted_state_sequence.shape[0]:
                predicted_state_sequence[k, :] = next_state

            current_state = next_state
        return (
            (predicted_state_sequence, action_sequence)
            if not return_predicted_states_only
            else predicted_state_sequence
        )


class EulerPredictor(Predictor):
    """Euler predictor uses a simple Euler discretization scheme.

    It does predictions by increments scaled by a sampling time times the velocity evaluated at each successive node.

    """

    def predict(self, current_state, action):
        next_state = (
            current_state
            + self.pred_step_size
            * self.system.compute_state_dynamics(
                time=None, state=current_state, inputs=action
            )
        )

        return next_state


class EulerPredictorMultistep(EulerPredictor):
    """Applies several iterations of Euler estimation to predict a single step."""

    def __init__(self, *args, n_steps=5, **kwargs):
        """Initialize an instance of EulerPredictorMultistep.

        Args:
            *args: positional arguments for EulerPredictor
            n_steps: number of estimations to predict a single step
            **kwargs: keyword arguments for EulerPredictor
        """
        super().__init__(*args, **kwargs)
        self.n_steps = n_steps
        self.pred_step_size /= self.n_steps

    def predict(self, current_state_or_observation, action):
        next_state_or_observation = current_state_or_observation
        for _ in range(self.n_steps):
            next_state_or_observation = super().predict(
                next_state_or_observation, action
            )
        return next_state_or_observation


class TrivialPredictor(Predictor):
    """A predictor that propagates the observation or state directly through the system dynamics law."""

    def __init__(self, system):
        """Initialize an instance of TrivialPredictor.

        Args:
            system: an instance of a discrete system
        """
        self.system = system

    def predict(self, time, state, action):
        return self.system.compute_state_dynamics(time, state, action)



"""The `entities` module of the optimization framework defines the core entities involved in setting up and managing the optimization problem.

This module contains the following key classes:

- `OptimizationVariable`: Represents a variable within the optimization process, which can be marked as a constant or a decision variable.
- `VarContainer`: A container class that holds and manages multiple `OptimizationVariable` instances, providing utility methods for variable manipulation.
- `NestedFunction`: A specialized form of `OptimizationVariable` that represents a nested function with its own set of dependent variables.
- `FunctionWithSignature`: A wrapper for functions within the optimization context, ensuring that the function signature matches the expected variables and managing constants and placeholders.
- `FuncContainer`: A container class that holds and manages multiple `FunctionWithSignature` instances, providing utility methods for function manipulation.
- `Hook`: A class representing a hook that can be applied to an `OptimizationVariable`.
- `ChainedHook`: A class representing a chain of hooks that can be applied to an `OptimizationVariable` one-by-one in a specific order.
These entities are designed to work together to define and manipulate the variables and functions that form the basis of an optimization problem or change it at runtime.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Callable, List, Tuple, Union, Optional, Dict
from typing_extensions import Self
from inspect import signature
from collections.abc import Mapping
import warnings
from .hooks import get_data_hook


@dataclass
class OptimizationVariable:
    """Represent an optimization variable within the optimization process.

    This base class encapsulates data and metadata for variables used in optimization,
    allowing them to be treated as either constants or decision variables.

    Args:
        name: The name of the variable.
        dims: The dimensions of the variable as a tuple.
        data: The initial data for the variable. Defaults to None.
        metadata: The metadata associated with the variable. Defaults to
            None.
        is_constant: Flag indicating if the variable is a constant.
            Defaults to False.
        hooks: A ChainedHook instance containing hooks to be applied to
            the variable. Defaults to an empty ChainedHook.
    """

    name: str
    dims: tuple
    data: Any = None
    metadata: Any = None
    is_constant: bool = False
    hooks: ChainedHook = field(default_factory=lambda: ChainedHook([]))

    def __call__(self, with_metadata: bool = False):
        """Retrieve the data or metadata of the variable after applying hooks.

        Args:
            with_metadata: If True, retrieve metadata; otherwise,
                retrieve data. Defaults to False.

        Returns:
            The data or metadata of the variable after hooks
            transformation.
        """
        obj_to_transform = self.data if not with_metadata else self.metadata
        hooks = self.hooks.metadata_hooks if with_metadata else self.hooks.data_hooks
        obj_to_transform = hooks(
            obj_to_transform, with_metadata=with_metadata, raw_eval=True
        )

        return obj_to_transform

    def renamed(self, new_name: str, inplace=True) -> Self:
        """Rename the optimization variable.

        Args:
            new_name: The new name for the variable.
            inplace: If True, modify the variable in place; otherwise,
                return a new instance. Defaults to True.

        Returns:
            The optimization variable with the updated name (self or new
            instance).
        """
        if inplace:
            self.name = new_name
            return self
        else:
            return OptimizationVariable(
                name=new_name,
                dims=self.dims,
                data=self.data,
                metadata=self.metadata,
                is_constant=self.is_constant,
            )

    def with_data(self, new_data, inplace=True):
        """Associate new data with the optimization variable.

        Args:
            new_data: The new data to associate with the variable.
            inplace: If True, modify the variable in place; otherwise,
                return a new instance. Defaults to True.

        Returns:
            The optimization variable with the updated data (self or new
            instance).
        """
        if inplace:
            self.data = new_data
            return self
        else:
            return OptimizationVariable(
                name=self.name,
                dims=self.dims,
                data=new_data,
                metadata=self.metadata,
                is_constant=self.is_constant,
            )

    def with_metadata(self, new_metadata, inplace=True):
        """Associate new metadata with the optimization variable.

        Args:
            new_metadata: The new metadata to associate with the
                variable.
            inplace: If True, modify the variable in place; otherwise,
                return a new instance. Defaults to True.

        Returns:
            The optimization variable with the updated metadata (self or
            new instance).
        """
        if inplace:
            self.metadata = new_metadata
            return self
        else:
            return OptimizationVariable(
                name=self.name,
                dims=self.dims,
                data=self.data,
                metadata=new_metadata,
                is_constant=self.is_constant,
            )

    def as_constant(self, inplace=True):
        """Mark the variable as a constant.

        Args:
            inplace: If True, modify the variable in place; otherwise,
                return a new instance marked as constant. Defaults to
                True.

        Returns:
            The optimization variable marked as a constant (self or new
            instance).
        """
        if self.is_constant:
            return self
        else:
            if inplace:
                self.is_constant = True
                return self
            else:
                return OptimizationVariable(
                    name=self.name,
                    dims=self.dims,
                    data=self.data,
                    metadata=self.metadata,
                    is_constant=True,
                )

    def as_decision_variable(self, inplace=True):
        """Mark the variable as a decision variable.

        Args:
            inplace: If True, modify the variable in place; otherwise,
                return a new instance marked as a decision variable.
                Defaults to True.

        Returns:
            The optimization variable marked as a decision variable
            (self or new instance).
        """
        if not self.is_constant:
            return self
        else:
            if inplace:
                self.is_constant = False
                return self
            return OptimizationVariable(
                name=self.name,
                dims=self.dims,
                data=self.data,
                metadata=self.metadata,
                is_constant=False,
            )

    def as_metadata_dict(self):
        """Convert the variable's metadata to a dictionary.

        Returns:
            A dictionary with the variable's name as the key and its
            metadata as the value.
        """
        return {self.name: self.metadata}

    def as_data_dict(self):
        """Convert the variable's data to a dictionary.

        Returns:
            A dictionary with the variable's name as the key and its
            data as the value.
        """
        return {self.name: self.data}

    def as_dims_dict(self):
        """Convert the variable's dimensions to a dictionary.

        Returns:
            A dictionary with the variable's name as the key and its
            dimensionsas the value.
        """
        return {self.name: self.dims}

    def __radd__(self, other):
        """Support the addition of this variable to another object by returning a VarContainer.

        Args:
            other: The object to add this variable to.

        Returns:
            A VarContainer containing this variable.
        """
        return VarContainer((self,))

    def __add__(self, other):
        """Add another variable or container to this variable to form a VarContainer.

        Args:
            other: The variable or VarContainer to be added.

        Returns:
            A VarContainer including this variable and `other`.
        """
        if isinstance(other, OptimizationVariable):
            return VarContainer([self, other])
        elif isinstance(other, VarContainer):
            return VarContainer((self,) + tuple(other.variables))

    def register_hook(self, hook: Union[Hook, Callable], first=False, act_on="all"):
        """Register a hook to the optimization variable.

        Args:
            hook: The hook or callable to be added as a hook.
            first: If True, the hook will be added at the beginning.
                Defaults to False.
            act_on: Specifies the target of the hook, either "data" or
                "metadata" or "all". Defaults to "all".
        """
        if not isinstance(hook, Hook):
            hook = Hook(hook, act_on=act_on)
        if self.hooks.hooks_container and hook.name in self.hooks.hooks_container.names:
            if first:
                hook_to_delete = [
                    hook_tmp
                    for hook_tmp in self.hooks.hooks_container
                    if hook_tmp.name == hook.name
                ][0]
                self.hooks.remove_hook(hook_to_delete)
                self.register_hook(hook, first=True)
            else:
                warnings.warn(f"Hook {hook.name} already registered.", stacklevel=1)
        else:
            self.hooks.register_hook(hook, first=first)

    def enable_hook(self, hook):
        """Enable a previously added hook.

        Args:
            hook: The hook to be enabled.
        """
        self.hooks.enable_hook(hook)

    def discard_hook(self, hook: FunctionWithSignature, disable_only=True):
        """Discard or disable a hook from the optimization variable.

        Args:
            hook: The hook to be discarded or disabled.
            disable_only: If True, the hook will only be disabled, not
                removed. Defaults to True.
        """
        if disable_only:
            self.hooks.disable_hook(hook)
        else:
            self.hooks.remove_hook(hook)

    def __str__(self):
        """Return a string representation of the optimization variable.

        Returns:
            A formatted string with the variable's details.
        """
        return (
            f"{self.name}\n  "
            + f"data: {self()}\n  "
            + f"metadata: {self(with_metadata=True)}\n  "
            + f"dims: {self.dims}\n  "
            + f"is_constant: {self.is_constant}\n\n"
        )


@dataclass
class VarContainer(Mapping):
    """A container class for managing multiple optimization variables.

    Args:
        _variables: A collection of OptimizationVariables to be
            contained.
    """

    _variables: Union[
        list[OptimizationVariable],
        Tuple[OptimizationVariable],
        Tuple[OptimizationVariable, OptimizationVariable],
    ]

    def __post_init__(self):
        """Initialize the VarContainer by ensuring variables are stored as a tuple."""
        self._variables = tuple(self._variables)

    @property
    def variables_hashmap(self):
        """Create a hashmap from variable names to variables.

        Returns:
            A dictionary mapping variable names to their respective
            OptimizationVariable objects.
        """
        return {var.name: var for var in self._variables}

    def to_dict(self):
        """Convert the container to a dictionary.

        Returns:
            A dictionary representation of the VarContainer.
        """
        return self.variables_hashmap

    def __repr__(self):
        """Return a string representation of the VarContainer for debugging.

        Returns:
            A debug string of the VarContainer.
        """
        return "VarContainer:\n  " + "".join(
            [f"{variable}\n  " for variable in self._variables]
        )

    def __str__(self):
        """Return a string representation of the VarContainer.

        Returns:
            A formatted string of the VarContainer.
        """
        return "VarContainer:\n  " + "".join(
            [f"{variable}\n  " for variable in self._variables]
        )

    @property
    def metadatas(self):
        """Retrieve the metadata of all contained variables.

        Returns:
            A tuple with the metadata of each OptimizationVariable.
        """
        return tuple(var.metadata for var in self.variables)

    def to_data_dict(self):
        """Convert variable data to a dictionary.

        Returns:
            A dictionary with variable names as keys and their data as
            values.
        """
        return {var.name: var() for var in self.variables}

    def to_metadata_dict(self):
        """Convert variable metadata to a dictionary.

        Returns:
            A dictionary with variable names as keys and their metadata
            as values.
        """
        return {var.name: var(with_metadata=True) for var in self.variables}

    def selected(self, var_names: List[str]) -> Tuple[OptimizationVariable]:
        """Select a subset of variables by names.

        Args:
            var_names: A list of names of the variables to select.

        Returns:
            A VarContainer containing only the selected variables.
        """
        return VarContainer([var for var in self.variables if var.name in var_names])

    def substitute_data(self, **name_data_dict) -> Self:
        """Substitute data for selected variables.

        Args:
            **name_data_dict: A dictionary with variable names as keys
                and new data as values.

        Returns:
            The current VarContainer with updated data for the selected
            variables.
        """
        for var in self.selected(list(name_data_dict.keys())):
            new_data = name_data_dict.get(var.name)
            var.with_data(new_data, inplace=True)
        return self

    def substitute_metadata(self, **name_metadata_dict) -> Self:
        """Substitute metadata for selected variables.

        Args:
            **name_metadata_dict: A dictionary with variable names as
                keys and new metadata as values.

        Returns:
            The current VarContainer with updated metadata for the
            selected variables.
        """
        for var in self.selected(list(name_metadata_dict.keys())):
            new_metadata = name_metadata_dict.get(var.name)
            var.with_metadata(new_metadata, inplace=True)
        return self

    @property
    def variables(
        self,
    ) -> Union[
        Tuple[OptimizationVariable],
        Tuple[OptimizationVariable, OptimizationVariable],
    ]:
        """Retrieve all variables contained within the VarContainer.

        Returns:
            A tuple of the contained OptimizationVariables.

        Raises:
            TypeError: If the internal variable storage is not a tuple
                or list.
        """
        if isinstance(self._variables, tuple):
            return self._variables
        elif isinstance(self._variables, list):
            return tuple(self._variables)
        else:
            raise TypeError("Something went wrong with the type of the variables.")

    @property
    def constants(self):
        """Retrieve all variables marked as constants from the container.

        Returns:
            A VarContainer containing only the variables marked as
            constants.
        """
        return VarContainer(
            tuple(variable for variable in self.variables if variable.is_constant)
        )

    @property
    def decision_variables(self):
        """Retrieve all variables not marked as constants from the container.

        Returns:
            A VarContainer containing only the variables not marked as
            constants.
        """
        return VarContainer(
            tuple(variable for variable in self.variables if not variable.is_constant)
        )

    @property
    def names(self):
        """Get the names of all variables contained within the VarContainer.

        Returns:
            A tuple containing the names of all variables.
        """
        return tuple(var.name for var in self.variables)

    def __radd__(self, other) -> Self:
        """Support the right-adding of this container to another object.

        Returns:
            The current VarContainer instance.
        """
        return self

    def __add__(self, other) -> Self:
        """Add another container or variable to this container.

        Args:
            other: The VarContainer or OptimizationVariable to be added.

        Returns:
            A new VarContainer instance containing the combined
            variables.

        Raises:
            NotImplementedError: If the addition operation is not
                defined for the `other` type.
        """
        if (
            isinstance(other, VarContainer)
            and isinstance(other.variables, tuple)
            and isinstance(self.variables, tuple)
        ):
            return VarContainer(other.variables + self.variables)
        elif isinstance(other, OptimizationVariable) and isinstance(
            self.variables, tuple
        ):
            return VarContainer(self.variables + (other,))
        elif other is None:
            return self
        else:
            raise NotImplementedError

    def __iter__(self):
        """Create an iterator over the contained variables.

        Returns:
            An iterator for the container's variables.
        """
        for variable in self.variables:
            yield variable

    def __len__(self):
        """Get the number of variables contained within the VarContainer.

        Returns:
            The number of contained variables.
        """
        return len(self._variables)

    def __getitem__(self, key) -> Union[OptimizationVariable, Self]:
        """Retrieve a contained variable by index, slice, or name.

        Args:
            key: The index, slice, or name used to retrieve the
                variable(s).

        Returns:
            The requested OptimizationVariable or a new VarContainer
            with the sliced variables.

        Raises:
            KeyError: If a variable with the given name is not found.
            NotImplementedError: If the key type is not supported.
        """
        if isinstance(key, int) and isinstance(self.variables, tuple):
            assert 0 <= key < len(self.variables), f"Index {key} is out of bounds."
            return self.variables[key]
        elif isinstance(key, slice) and isinstance(self.variables, tuple):
            return VarContainer(self.variables[key])
        elif isinstance(key, str) and self.variables_hashmap is not None:
            res = self.variables_hashmap.get(key)
            if res is None:
                raise KeyError(f"OptimizationVariable {key} not found.")
            return res
        else:
            raise NotImplementedError

    def fix(
        self,
        variables_to_fix: list[str],
        hook: Optional[Hook] = None,
    ) -> None:
        """Fix the values of the specified variables and optionally apply a hook.

        Args:
            variables_to_fix: A list of variable names to be fixed.
            hook: An optional Hook object to be applied to the fixed
                variables.
        """
        for var in self.decision_variables.selected(variables_to_fix):
            var.as_constant(inplace=True)
            hook_names = [hook.name for hook in var.hooks]
            if "requires_grad" in hook_names:
                req_grad_hook = [
                    hook for hook in var.hooks if "requires_grad" in hook.name
                ][0]
                var.discard_hook(hook=req_grad_hook)
            if hook:
                assert (
                    hook.name == "detach"
                ), f"Unfixing of variable should be provided with detach hook, not {hook.name}"

                var.register_hook(hook=get_data_hook(var=var), first=True)
                var.register_hook(hook=hook, first=True)

    def unfix(
        self,
        variables_to_unfix: list[str],
        hook: Optional[Hook] = None,
    ) -> None:
        """Unfix the values of the specified variables and optionally apply a hook.

        Args:
            variables_to_unfix: A list of variable names to be unfixed.
            hook: An optional Hook object to be applied to the unfixed
                variables.
        """
        for var in self.constants.selected(variables_to_unfix):
            var.as_decision_variable(inplace=True)
            hook_names = [hook.name for hook in var.hooks]
            if "detach" in hook_names:
                detach_hook = [hook for hook in var.hooks if "detach" in hook.name][0]
                var.discard_hook(hook=detach_hook)
            if hook:
                assert (
                    hook.name == "requires_grad"
                ), f"Unfixing of variable should be provided with requires_grad hook, not {hook.name}"

                var.register_hook(hook=get_data_hook(var=var), first=True)
                var.register_hook(hook=hook, first=True)


@dataclass
class NestedFunction(OptimizationVariable):
    """Represent a nested function within the optimization process.

    This class extends the OptimizationVariable to represent a nested function, including variables on which it depends.

    Args:
        nested_variables: A container for variables that the nested
            function depends on. Defaults to an empty VarContainer.
    """

    nested_variables: VarContainer = field(default_factory=lambda: VarContainer([]))

    def substitute_parameters(self, **parameters):
        """Substitute parameters within the nested variables.

        Args:
            **parameters: A dictionary where keys are variable names and
                values are the new data to substitute.
        """
        self.nested_variables.substitute_data(**parameters)
        self.nested_variables.substitute_metadata(**parameters)

    def with_data(self, new_data, inplace=True):
        """Associate new data with the nested function.

        Args:
            new_data: The new data to associate with the nested
                function.
            inplace: If True, modify the nested function in place;
                otherwise, return a new instance. Defaults to True.

        Returns:
            The nested function with the updated data (self or new
            instance).
        """
        if inplace:
            self.data = new_data
            return self
        else:
            return NestedFunction(
                name=self.name,
                dims=self.dims,
                data=new_data,
                metadata=self.metadata,
                is_constant=self.is_constant,
                nested_variables=self.nested_variables,
            )

    def with_metadata(self, new_metadata, inplace=True):
        """Associate new metadata with the nested function.

        Args:
            new_metadata: The new metadata to associate with the nested
                function.
            inplace: If True, modify the nested function in place;
                otherwise, return a new instance. Defaults to True.

        Returns:
            The nested function with the updated metadata (self or new
            instance).
        """
        if inplace:
            self.metadata = new_metadata
            return self
        else:
            return NestedFunction(
                name=self.name,
                dims=self.dims,
                data=self.data,
                metadata=new_metadata,
                is_constant=self.is_constant,
                nested_variables=self.nested_variables,
            )

    def renamed(self, new_name: str, inplace=True) -> Self:
        """Rename the nested function.

        Args:
            new_name: The new name for the nested function.
            inplace: If True, modify the nested function in place;
                otherwise, return a new instance. Defaults to True.

        Returns:
            The nested function with the updated name (self or new
            instance).
        """
        if inplace:
            self.name = new_name
            return self
        else:
            return NestedFunction(
                name=new_name,
                dims=self.dims,
                data=self.data,
                metadata=self.metadata,
                is_constant=self.is_constant,
                nested_variables=self.nested_variables,
            )

    def as_constant(self, inplace=True):
        """Mark the nested function as a constant.

        Args:
            inplace: If True, modify the nested function in place;
                otherwise, return a new instance marked as constant.
                Defaults to True.

        Returns:
            The nested function marked as a constant (self or new
            instance).
        """
        if self.is_constant:
            return self
        else:
            if inplace:
                self.is_constant = True
                return self
            else:
                return NestedFunction(
                    name=self.name,
                    dims=self.dims,
                    data=self.data,
                    metadata=self.metadata,
                    is_constant=True,
                    nested_variables=self.nested_variables,
                )

    def as_decision_variable(self, inplace=True):
        """Mark the nested function as a decision variable.

        Args:
            inplace: If True, modify the nested function in place;
                otherwise, return a new instance marked as a decision
                variable. Defaults to True.

        Returns:
            The nested function marked as a decision variable (self or
            new instance).
        """
        if not self.is_constant:
            return self
        else:
            if inplace:
                self.is_constant = False
                return self
            return NestedFunction(
                name=self.name,
                dims=self.dims,
                data=self.data,
                metadata=self.metadata,
                is_constant=False,
                nested_variables=self.nested_variables,
            )


@dataclass
class FunctionWithSignature:
    """Wraps a callable function and manage its signature for the optimization process.

    This class parses the function's signature to ensure that it complies with the expected format for the optimization procedure.

    Args:
        func: The callable function to be wrapped.
        variables: A container of variables that the function operates
            on. Defaults to an empty VarContainer.
        is_objective: Flag indicating if the function is an objective
            function. Defaults to False.
        metadata: Additional metadata for the function. Defaults to
            None.
    """

    func: Callable
    variables: VarContainer = field(default_factory=lambda: VarContainer([]))
    is_objective: bool = False
    metadata: Any = None

    def __post_init__(self) -> None:
        """Initialize the function wrapper by parsing its signature.

        Raises:
            ValueError: If there are unknown variables in the signature
                that are not in the variable container.
        """
        self.__signature = self.__parse_signature(self.func)
        parameter_names = set(self.variables.names)
        kwargs_intersection = parameter_names & set(self.__signature)
        if kwargs_intersection != parameter_names:
            raise ValueError(
                "Unknown variables encountered: "
                f"{self.variables.keys() - kwargs_intersection}"
            )

        self.name = self.func.__name__

    @property
    def await_constants(self):
        """Determine if there are constants awaiting substitution.

        Returns:
            True if there are constants to substitute, False otherwise.
        """
        return not (self.constants_to_substitute == [])

    @property
    def constants_to_substitute(self):
        """List the names of constants that need to be substituted.

        Returns:
            A list of constant names that require substitution.
        """
        return [var.name for var in self.variables.constants if var() is None]

    def __call__(
        self, *args, with_metadata: bool = False, raw_eval: bool = False, **kwargs
    ) -> Any:
        """Call the wrapped function with the provided arguments and keyword arguments.

        Args:
            *args: Positional arguments to be passed to the function.
            with_metadata: If True, pass metadata to the function.
                Defaults to False.
            raw_eval: If True, evaluate the function with raw data.
                Defaults to False.
            **kwargs (Dict[str, Any]) : Keyword arguments to be passed to the function.

        Returns:
            The result of the function call.

        Raises:
            ValueError: If not all required constants have been
                substituted or if excess arguments are provided.
        """
        if raw_eval:
            kwargs_to_pass = {
                k: v for k, v in kwargs.items() if k in self.variables.names
            }
            return self.func(*args, **kwargs_to_pass)

        if kwargs == {} and len(args) == 1:
            dvar = self.variables[self.free_placeholders[0]]
            is_decision_var_nested_function = isinstance(dvar, NestedFunction)
            if is_decision_var_nested_function:
                assert (
                    len(self.free_placeholders) == 1
                ), "The amount of free placeholders should be 1"
                dvar.substitute_parameters(
                    **{dvar.nested_variables.decision_variables.names[0]: args[0]}
                )
            else:
                dvar.with_data(args[0])
            return self.func(
                **{self.free_placeholders[0]: dvar()},
                **{
                    k: v
                    for k, v in self.variables.to_data_dict().items()
                    if k != self.free_placeholders[0]
                },
            )

        kwargs_to_pass = {
            k: v for k, v in kwargs.items() if k in self.free_placeholders
        }

        if with_metadata:
            return self.func(
                **{**kwargs_to_pass, **self.variables.constants.to_metadata_dict()}
            )
        elif not self.await_constants:
            return self.func(
                **{
                    **kwargs_to_pass,
                    **self.variables.constants.to_data_dict(),
                }
            )
        else:
            raise ValueError(
                f"Not all declared constants were substituted: {self.constants_to_substitute}"
            )

    @property
    def signature(self) -> tuple:
        """Get the signature of the wrapped function.

        Returns:
            A tuple representing the function's signature.
        """
        return tuple(self.__signature)

    @property
    def occupied(self) -> tuple:
        """Get the names of constants in the variables.

        Returns:
            A tuple containing the names of all constant variables.
        """
        return tuple(self.variables.constants.names)

    @property
    def free_placeholders(self) -> tuple:
        """Returns a list of free variables of the current function.

        Free variables are the arguments that
        are not defaulted and do not have a corresponding value.
        This method uses the signature of the function
        and the default parameters keys to determine the free variables.

        Returns:
            tuple: A list of free variables of the current function.
        """
        signature_set = set(self.__signature)
        # default_keys = {
        #     name: data
        #     for name, data in self.variables.to_data_dict().items()
        #     if data is not None
        # }.keys()
        default_keys = set(self.variables.constants.names)
        return tuple(signature_set - default_keys)

    def declare_variables(
        self,
        variables: Union[VarContainer, OptimizationVariable],
        replace: bool = False,
    ) -> Self:
        """Declare variables for the function, optionally replacing existing ones.

        Args:
            variables: The variables to declare for the function.
            replace: If True, replace any existing variables. Defaults
                to False.

        Returns:
            The instance of the function with declared variables.

        Raises:
            AssertionError: If unknown parameters are provided.
        """
        if isinstance(variables, OptimizationVariable):
            variables = VarContainer([variables])
        unknown_params = [
            name for name in variables.names if name not in self.__signature
        ]
        assert unknown_params == [], f"Unknown parameters: {unknown_params}"

        if not replace:
            if self.variables is not None:
                new_variables = self.variables + variables
                if new_variables is not None:
                    self.variables = self.variables + variables
        else:
            self.variables = variables
        return self

    def set_parameters(self, **kwargs) -> None:
        """Set parameters for the function.

        Args:
            **kwargs (Dict[str, Any]): Keyword arguments representing the parameters to
                set.

        Raises:
            ValueError: If unknown parameters are encountered.
        """
        kwargs_intersection = kwargs.keys() & self.occupied
        if kwargs_intersection != kwargs.keys():
            raise ValueError(
                f"Unknown parameters encountered: {kwargs.keys() - kwargs_intersection}"
            )
        self.variables.substitute_data(**kwargs)

    def fix_variables(
        self,
        variables_to_fix: List[str],
        data_dict: Optional[Dict],
        metadata_dict: Optional[Dict],
        hook: Optional[Self] = None,
    ):
        """Fix variables and optionally set their data and metadata.

        Args:
            variables_to_fix: A list of variable names to be fixed.
            data_dict: Optional dictionary with data to set for fixed
                variables.
            metadata_dict: Optional dictionary with metadata to set for
                fixed variables.
            hook: Optional hook to be applied to fixed variables.
        """
        self.variables.fix(variables_to_fix, hook=hook)
        if data_dict:
            self.set_parameters(**data_dict)
        if metadata_dict:
            self.variables.substitute_metadata(**metadata_dict)

    def unfix_variables(
        self,
        variables_to_unfix: List[str],
        hook: Optional[Self] = None,
    ):
        """Unfix variables and optionally apply a hook.

        Args:
            variables_to_unfix: A list of variable names to be unfixed.
            hook: Optional hook to be applied to unfixed variables.
        """
        self.variables.unfix(variables_to_unfix, hook=hook)

    def __parse_signature(self, func: Callable) -> Tuple[str]:
        """Parse the signature of the given function.

        Args:
            func: The function to parse the signature for.

        Returns:
            A tuple representing the parsed signature.

        Raises:
            ValueError: If the function has undefined positional or
                keyword arguments.
        """
        signature_list = []

        variables = (
            func.signature
            if hasattr(func, "signature")
            else signature(func).parameters.values()
        )
        for param in variables:
            if param.kind == param.VAR_POSITIONAL:
                raise ValueError("Undefined number of arguments")
            if param.kind == param.VAR_KEYWORD:
                raise ValueError("Undefined number of keyword arguments")
            signature_list.append(param.name)

        return tuple(signature_list)

    def __radd__(self, other: Self) -> FuncContainer:
        """Support the right-adding of this function to another object to form a FuncContainer.

        Args:
            other: The object to add this function to.

        Returns:
            A FuncContainer containing this function.
        """
        return FuncContainer((self,))

    def __add__(self, other: Self) -> Optional[FuncContainer]:
        """Add another function or container to this function to form a FuncContainer.

        Args:
            other: The function or FuncContainer to be added.

        Returns:
            A FuncContainer including this function and `other`, or None
            if `other` is not of a supported type.
        """
        if isinstance(other, FunctionWithSignature):
            return FuncContainer((self, other))
        elif isinstance(other, FuncContainer):
            return FuncContainer((self, *other.functions))
        else:
            return None


@dataclass
class FuncContainer(Mapping):
    """A container for functions with signature."""

    _functions: Union[
        Tuple[FunctionWithSignature, FunctionWithSignature],
        Tuple[FunctionWithSignature],
    ]

    def __post_init__(self):
        self._functions = tuple(self._functions)

    @property
    def functions_hashmap(self):
        return {func.name: func for func in self._functions}

    def to_dict(self):
        return self.functions_hashmap

    def __repr__(self):
        return "FuncContainer:\n  " + "".join(
            [f"{function}\n  " for function in self._functions]
        )

    def __str__(self):
        return "FuncContainer:\n  " + "".join(
            [f"{function}\n  " for function in self._functions]
        )

    @property
    def functions(self):
        return self._functions

    @property
    def objectives(self):
        return FuncContainer(
            tuple(function for function in self.functions if function.is_objective)
        )

    @property
    def constraints(self):
        return FuncContainer(
            tuple(function for function in self.functions if not function.is_objective)
        )

    def __radd__(self, other):
        return self

    def __add__(self, other):
        if isinstance(other, FuncContainer):
            return FuncContainer(other.functions + self.functions)
        elif isinstance(other, FunctionWithSignature):
            return FuncContainer(self.functions + (other,))

    def __iter__(self):
        for function in self.functions:
            yield function

    def __len__(self):
        return len(self.functions)

    def __getitem__(self, key) -> Union[FunctionWithSignature, Self]:
        if isinstance(key, int):
            assert 0 <= key < len(self.functions), f"Index {key} is out of bounds."
            return self.functions[key]
        elif isinstance(key, slice):
            return FuncContainer(self.functions[key])
        elif isinstance(key, str) and self.functions_hashmap is not None:
            res = self.functions_hashmap.get(key)
            assert res is not None, f"Function {key} not found."
            return res
        else:
            raise NotImplementedError

    @property
    def names(self):
        return tuple(function.name for function in self.functions)


@dataclass
class Hook(FunctionWithSignature):
    """Base class for all hooks.

    Hooks are to be used with `OptimizationVariable` objects in order to modify `FunctionWithSignature` calling.
    """

    act_on: str = "all"


class ChainedHook:
    """A container for hooks."""

    def __init__(self, hooks: List[Hook]) -> None:
        """Initialize a chained hook object."""
        self.hooks = hooks
        self.hooks_container = sum(self.hooks) if len(hooks) > 0 else FuncContainer([])
        self.enabled_hashmap = {hook.name: True for hook in hooks}

    def __call__(self, arg, **kwargs):
        curr_result = arg
        for hook in [hook for hook in self.hooks if self.enabled_hashmap[hook.name]]:
            curr_result = self.call_hook(hook.name, curr_result, **kwargs)
        return curr_result

    def permute_order(self, hook1, hook2):
        for i, hook in enumerate(self.hooks):
            if hook == hook1:
                self.hooks[i] = hook2
            elif hook == hook2:
                self.hooks[i] = hook1

    def set_priority(self, hook1, hook2):
        assert (
            hook1 in self.hooks and hook2 in self.hooks
        ), f"Hook {hook1} or {hook2} not found."
        if self.hooks.index(hook1) > self.hooks.index(hook2):
            self.permute_order(hook1, hook2)

    def __iter__(self):
        for hook in self.hooks:
            yield hook

    def __len__(self):
        return len(self.hooks)

    def register_hook(self, hook, first=True):
        if hook not in self.hooks:
            self.__append_hook(hook, first=first)
        else:
            self.enabled_hashmap[hook.name] = True

    def enable_hook(self, hook):
        assert hook in self.hooks, f"Hook {hook} not found."
        self.enabled_hashmap[hook.name] = True

    def __append_hook(self, hook, first=True):
        if first:
            self.hooks.insert(0, hook)
        else:
            self.hooks.append(hook)

        self.hooks_container = (
            sum(self.hooks) if len(self.hooks) > 1 else FuncContainer(tuple(self.hooks))
        )
        self.enabled_hashmap[hook.name] = True

    def disable_hook(self, hook):
        assert hook in self.hooks, f"Hook {hook} not found."
        self.enabled_hashmap[hook.name] = False

    def remove_hook(self, hook):
        assert hook in self.hooks, f"Hook {hook} not found."
        self.hooks.remove(hook)
        del self.enabled_hashmap[hook.name]
        self.hooks_container = sum(self.hooks)

    def call_hook(self, name, *args, with_metadata=False, **kwargs):
        assert name in self.enabled_hashmap.keys(), f"Hook {name} not found."
        assert isinstance(
            self.hooks_container, FuncContainer
        ), f"Unexpected type of hook container: {type(self.hooks_container)}."
        hook = self.hooks_container[name]
        if isinstance(hook, FunctionWithSignature):
            return hook(*args, **kwargs, with_metadata=with_metadata)
        else:
            raise NotImplementedError

    @property
    def metadata_hooks(self):
        return ChainedHook(
            [
                hook
                for hook in self.hooks_container
                if (hook.act_on == "metadata" or hook.act_on == "all")
                and self.enabled_hashmap[hook.name]
            ]
        )

    @property
    def data_hooks(self):
        return ChainedHook(
            [
                hook
                for hook in self.hooks_container
                if (hook.act_on == "data" or hook.act_on == "all")
                and self.enabled_hashmap[hook.name]
            ]
        )
